"""Base class of PhonopyCalculation and Phono3pyCalculation."""

from aiida.engine import CalcJob
from aiida.common import CalcInfo, CodeInfo
from aiida.plugins import DataFactory
from aiida_phonopy.common.file_generators import get_BORN_txt

Dict = DataFactory("dict")
StructureData = DataFactory("structure")
ArrayData = DataFactory("array")
Float = DataFactory("float")
Bool = DataFactory("bool")


class BasePhonopyCalculation(CalcJob):
    """A basic plugin for calculating force constants using Phonopy.

    Requirement: the node should be able to import phonopy if NAC is used

    """

    _INPUT_NAC = "BORN"

    @classmethod
    def define(cls, spec):
        """Define inputs, outputs, and outline."""
        super().define(spec)
        spec.input("settings", valid_type=Dict, help="Phonopy parameters.")
        spec.input("structure", valid_type=StructureData, help="Unit cell structure.")
        spec.input("symmetry_tolerance", valid_type=Float, default=lambda: Float(1e-5))
        spec.input(
            "fc_only",
            valid_type=Bool,
            help="Only force constants are calculated.",
            default=lambda: Bool(False),
        )
        spec.input(
            "force_sets",
            valid_type=ArrayData,
            required=False,
            help="Sets of forces in supercells",
        )
        spec.input(
            "nac_params", valid_type=ArrayData, required=False, help="NAC parameters."
        )
        spec.input(
            "primitive",
            valid_type=StructureData,
            required=False,
            help="Primitive cell structure only necessary NAC is applied.",
        )
        spec.input(
            "dataset",
            valid_type=(Dict, ArrayData),
            required=False,
            help="Displacements and forces dataset.",
        )

    def prepare_for_submission(self, folder):
        """Prepare calcinfo."""
        self.logger.info("prepare_for_submission")

        # These three lists are updated in self._create_additional_files(folder)
        self._internal_retrieve_list = []
        self._additional_cmd_params = []
        self._calculation_cmd = []
        self._create_additional_files(folder)

        # ================= prepare the python input files =================

        # BORN
        if (
            not self.inputs.fc_only
            and "nac_params" in self.inputs
            and "primitive" in self.inputs
        ):
            born_txt = get_BORN_txt(
                self.inputs.nac_params,
                self.inputs.primitive,
                self.inputs.symmetry_tolerance,
            )
            with folder.open(self._INPUT_NAC, "w", encoding="utf8") as handle:
                handle.write(born_txt)
            for params in self._additional_cmd_params:
                params.append("--nac")

        self.inputs.metadata.options.withmpi = False

        # ============================ calcinfo ===============================

        local_copy_list = []
        remote_copy_list = []

        calcinfo = CalcInfo()
        calcinfo.uuid = self.uuid
        calcinfo.local_copy_list = local_copy_list
        calcinfo.remote_copy_list = remote_copy_list
        calcinfo.retrieve_list = self._internal_retrieve_list

        calcinfo.codes_info = []
        for i, (default_params, additional_params) in enumerate(
            zip(self._calculation_cmd, self._additional_cmd_params)
        ):
            codeinfo = CodeInfo()
            cmdline_params = default_params + additional_params
            codeinfo.cmdline_params = cmdline_params
            codeinfo.code_uuid = self.inputs.code.uuid
            codeinfo.withmpi = False
            calcinfo.codes_info.append(codeinfo)

        return calcinfo

    def _create_additional_files(self, folder):
        raise NotImplementedError()



"""WorkChan to run ph-ph calculation by phono3py and force calculators."""

from aiida.engine import WorkChain, calcfunction, if_
from aiida.orm import Bool, Code, Float
from aiida.plugins import DataFactory, WorkflowFactory

from aiida_phonopy.common.builders import (
    get_vasp_immigrant_inputs,
    get_workchain_inputs,
)
from aiida_phonopy.common.utils import (
    collect_vasp_forces_and_energies,
    compare_structures,
    generate_phono3py_cells,
    get_vasp_force_sets_dict,
)
from aiida_phonopy.workflows.nac_params import _get_nac_params as get_nac_params

PhonopyWorkChain = WorkflowFactory("phonopy.phonopy")
Dict = DataFactory("dict")
ArrayData = DataFactory("array")
StructureData = DataFactory("structure")


def get_calcjob_builder(structure, code_string, builder_inputs, label=None):
    """Return process builder.

    This method supposes createing a process builder of a force calculator
    (VASP, QE, etc.).

    """
    code = Code.get_from_string(code_string)
    if code.get_input_plugin_name() == "vasp.vasp":
        VaspWorkflow = WorkflowFactory("vasp.vasp")
        builder = VaspWorkflow.get_builder()
        if label:
            builder.metadata.label = label
        builder.code = Code.get_from_string(code_string)
        builder.structure = structure
        builder.settings = builder_inputs["settings"]
        builder.parameters = builder_inputs["parameters"]
        builder.kpoints = builder_inputs["kpoints"]
        builder.potential_family = builder_inputs["potential_family"]
        builder.potential_mapping = builder_inputs["potential_mapping"]
        builder.options = builder_inputs["options"]
        builder.clean_workdir = Bool(False)
    else:
        raise RuntimeError("Code could not be found.")

    return builder


@calcfunction
def get_force_calcjob_inputs(calculator_settings, supercell):
    """Return builder inputs of force calculations."""
    return get_workchain_inputs(calculator_settings, supercell, calc_type="forces")


@calcfunction
def get_phonon_force_calcjob_inputs(calculator_settings, supercell):
    """Return builder inputs of force calculations for phono3py fc2."""
    return get_workchain_inputs(
        calculator_settings, supercell, calc_type="phonon_forces"
    )


@calcfunction
def get_nac_calcjob_inputs(calculator_settings, unitcell):
    """Return builder inputs of an NAC params calculation."""
    return get_workchain_inputs(calculator_settings, unitcell, "nac")


class Phono3pyWorkChain(WorkChain):
    """Phono3py workchain."""

    @classmethod
    def define(cls, spec):
        """Define inputs, outputs, and outline."""
        super().define(spec)
        spec.expose_inputs(
            PhonopyWorkChain,
            exclude=["calculation_nodes", "run_phonopy", "remote_phonopy"],
        )
        spec.input(
            "run_phono3py", valid_type=Bool, required=False, default=lambda: Bool(False)
        )
        spec.input(
            "remote_phono3py",
            valid_type=Bool,
            required=False,
            default=lambda: Bool(False),
        )

        spec.outline(
            cls.initialize,
            if_(cls.import_calculations_from_files)(
                cls.read_force_and_nac_calculations_from_files,
                cls.check_imported_structures,
            ).else_(
                cls.run_force_and_nac_calculations,
            ),
            if_(cls.dry_run)(cls.postprocess_of_dry_run,).else_(
                cls.create_force_sets,
                if_(cls.is_nac)(cls.create_nac_params),
                if_(cls.run_phono3py)(
                    if_(cls.remote_phono3py)(
                        cls.run_phono3py_remote,
                        cls.collect_data,
                    ).else_(
                        cls.create_force_constants,
                        cls.run_phono3py_in_workchain,
                    )
                ),
            ),
        )
        spec.output("fc3", valid_type=ArrayData, required=False)
        spec.output("fc2", valid_type=ArrayData, required=False)
        spec.output("primitive", valid_type=StructureData, required=False)
        spec.output("supercell", valid_type=StructureData, required=False)
        spec.output("phonon_supercell", valid_type=StructureData, required=False)
        spec.output("force_sets", valid_type=ArrayData, required=False)
        spec.output("supercell_forces", valid_type=ArrayData, required=False)
        spec.output("supercell_energy", valid_type=Float, required=False)
        spec.output("phonon_force_sets", valid_type=ArrayData, required=False)
        spec.output("phonon_supercell_forces", valid_type=ArrayData, required=False)
        spec.output("phonon_supercell_energy", valid_type=Float, required=False)
        spec.output("nac_params", valid_type=ArrayData, required=False)
        spec.output("phonon_setting_info", valid_type=Dict, required=True)
        spec.exit_code(
            700,
            "ERROR_NO_FORCES_CALCULATOR_SETTING",
            message="Force calculator setting not found.",
        )
        spec.exit_code(
            701,
            "ERROR_NO_NAC_CALCULATOR_SETTING",
            message="NAC calculator setting not found.",
        )
        spec.exit_code(
            710,
            "ERROR_NO_FORCES_FOLDERS",
            message="Forces calculation folders for immigrant not found.",
        )
        spec.exit_code(
            711,
            "ERROR_NO_NAC_FOLDER",
            message="NAC calculation folder for immigrant not found.",
        )

    def is_nac(self):
        if "is_nac" in self.inputs.phonon_settings.attributes:
            return self.inputs.phonon_settings["is_nac"]
        else:
            False

    def dry_run(self):
        return self.inputs.dry_run

    def remote_phono3py(self):
        return self.inputs.remote_phono3py

    def run_phono3py(self):
        return self.inputs.run_phono3py

    def import_calculations_from_files(self):
        return "immigrant_calculation_folders" in self.inputs

    def initialize(self):
        """Set default settings and create supercells and primitive cell"""

        self.report("initialize")

        if "forces" not in self.inputs.calculator_settings.attributes:
            self.report("'forces' item is necessary in " "calculator_settings input.")
            return self.exit_code.ERROR_NO_FORCES_CALCULATOR_SETTING

        if self.is_nac():
            if "nac" not in self.inputs.calculator_settings.attributes:
                self.report("'nac' item is necessary in " "calculator_settings input.")
                return self.exit_code.ERROR_NO_NAC_CALCULATOR_SETTING

        if self.inputs.run_phono3py and self.inputs.remote_phono3py:
            if "code_string" not in self.inputs or "options" not in self.inputs:
                raise RuntimeError("code_string and options have to be specified.")

        if "supercell_matrix" not in self.inputs.phonon_settings.attributes:
            raise RuntimeError("supercell_matrix was not found in phonon_settings.")

        kwargs = {}
        if "displacement_dataset" in self.inputs:
            kwargs["dataset"] = self.inputs.displacement_dataset
        return_vals = generate_phono3py_cells(
            self.inputs.phonon_settings,
            self.inputs.structure,
            self.inputs.symmetry_tolerance,
            **kwargs
        )

        for key in ("phonon_setting_info", "primitive", "supercell"):
            self.ctx[key] = return_vals[key]
            self.out(key, self.ctx[key])
        self.ctx.supercells = {}
        self.ctx.phonon_supercells = {}
        for key in return_vals:
            if "supercell_" in key and "phonon_" not in key:
                self.ctx.supercells[key] = return_vals[key]
            if "phonon_supercell_" in key:
                self.ctx.phonon_supercells[key] = return_vals[key]
        self.ctx.primitive = return_vals["primitive"]
        self.ctx.supercell = return_vals["supercell"]
        if "phonon_supercell" in return_vals:
            self.ctx.phonon_supercell = return_vals["phonon_supercell"]

        if self.inputs.subtract_residual_forces:
            digits = len(str(len(self.ctx.supercells)))
            label = "supercell_%s" % "0".zfill(digits)
            self.ctx.supercells[label] = return_vals["supercell"]
            digits = len(str(len(self.ctx.phonon_supercells)))
            label = "phonon_supercell_%s" % "0".zfill(digits)
            self.ctx.phonon_supercells[label] = return_vals["phonon_supercell"]

    def postprocess_of_dry_run(self):
        self.report("Finish here because of dry-run setting")

    def run_force_and_nac_calculations(self):
        self._run_force_calculations()
        if "phonon_supercell" in self.ctx:
            self._run_phonon_force_calculations()
        if self.is_nac():
            self._run_nac_calculation()

    def _run_force_calculations(self):
        """FC3 force calculation"""
        self.report("run force calculations")
        builder_inputs = get_force_calcjob_inputs(
            self.inputs.calculator_settings, self.ctx.supercell
        )
        for key in self.ctx.supercells:
            builder = get_calcjob_builder(
                self.ctx.supercells[key],
                self.inputs.calculator_settings["forces"]["code_string"],
                builder_inputs,
                label=key,
            )
            future = self.submit(builder)
            label = "force_calc_%s" % key.split("_")[-1]
            self.report("{} pk = {}".format(label, future.pk))
            self.to_context(**{label: future})

    def _run_phonon_force_calculations(self):
        """FC2 force calculation"""
        self.report("run phonon force calculations")
        calc_settings = self.inputs.calculator_settings
        builder_inputs = get_phonon_force_calcjob_inputs(
            calc_settings, self.ctx.phonon_supercell
        )
        for key in self.ctx.phonon_supercells:
            builder = get_calcjob_builder(
                self.ctx.phonon_supercells[key],
                calc_settings["phonon_forces"]["code_string"],
                builder_inputs,
                label=key,
            )
            future = self.submit(builder)
            label = "phonon_force_calc_%s" % key.split("_")[-1]
            self.report("{} pk = {}".format(label, future.pk))
            self.to_context(**{label: future})

    def _run_nac_calculation(self):
        """Born charges and dielectric constant calculation"""
        self.report("calculate born charges and dielectric constant")
        builder_inputs = get_nac_calcjob_inputs(
            self.inputs.calculator_settings, self.ctx.primitive
        )
        builder = get_calcjob_builder(
            self.ctx.primitive,
            self.inputs.calculator_settings["nac"]["code_string"],
            builder_inputs,
            label="born_and_epsilon",
        )
        future = self.submit(builder)
        self.report("born_and_epsilon: {}".format(future.pk))
        self.to_context(**{"born_and_epsilon_calc": future})

    def read_force_and_nac_calculations_from_files(self):
        self.report("import calculation data in files")

        calc_folders_Dict = self.inputs.immigrant_calculation_folders
        VaspImmigrant = WorkflowFactory("vasp.immigrant")

        if "forces" not in calc_folders_Dict.attributes:
            return self.exit_code.ERROR_NO_FORCES_FOLDERS

        if self.is_nac():  # NAC the last one
            if "nac" not in calc_folders_Dict.attributes:
                return self.exit_code.ERROR_NO_NAC_FOLDER

        digits = len(str(len(calc_folders_Dict["forces"])))
        for i, force_folder in enumerate(calc_folders_Dict["forces"]):
            label = "force_calc_%s" % str(i + 1).zfill(digits)
            inputs = get_vasp_immigrant_inputs(
                force_folder, self.inputs.calculator_settings["forces"], label=label
            )
            future = self.submit(VaspImmigrant, **inputs)
            self.report("{} pk = {}".format(label, future.pk))
            self.to_context(**{label: future})

        if "phonon_forces" in calc_folders_Dict.attributes:
            folders = calc_folders_Dict["phonon_forces"]
            digits = len(str(len(folders)))
            for i, force_folder in enumerate(folders):
                label = "force_calc_%s" % str(i + 1).zfill(digits)
                inputs = get_vasp_immigrant_inputs(
                    force_folder,
                    self.inputs.calculator_settings["phonon_forces"],
                    label=label,
                )
                future = self.submit(VaspImmigrant, **inputs)
                self.report("{} pk = {}".format(label, future.pk))
                self.to_context(**{label: future})

        if self.is_nac():  # NAC the last one
            label = "born_and_epsilon_calc"
            inputs = get_vasp_immigrant_inputs(
                calc_folders_Dict["nac"][0],
                self.inputs.calculator_settings["nac"],
                label=label,
            )
            future = self.submit(VaspImmigrant, **inputs)
            self.report("{} pk = {}".format(label, future.pk))
            self.to_context(**{label: future})

    def check_imported_structures(self):
        self.report("check imported supercell structures")

        msg = "Immigrant failed because of inconsistency of supercell" "structure"

        for key in self.ctx.supercells:
            num = key.split("_")[-1]
            calc = self.ctx["force_calc_%s" % num]
            if type(calc) is dict:  # dict when using get_data_from_node_id
                calc_dict = calc
            else:
                calc_dict = calc.inputs
            supercell_ref = self.ctx.supercells["supercell_%s" % num]
            supercell_calc = calc_dict["structure"]
            if not compare_structures(
                supercell_ref, supercell_calc, self.inputs.symmetry_tolerance
            ):
                raise RuntimeError(msg)

        for key in self.ctx.phonon_supercells:
            num = key.split("_")[-1]
            calc = self.ctx["phonon_force_calc_%s" % num]
            if type(calc) is dict:  # dict when using get_data_from_node_id
                calc_dict = calc
            else:
                calc_dict = calc.inputs
            supercell_ref = self.ctx.phonon_supercells["phonon_supercell_%s" % num]
            supercell_calc = calc_dict["structure"]
            if not compare_structures(
                supercell_ref, supercell_calc, self.inputs.symmetry_tolerance
            ):
                raise RuntimeError(msg)

    def create_force_sets(self):
        """Build datasets from forces of supercells with displacments"""

        self.report("create force sets")

        forces_dict = collect_vasp_forces_and_energies(self.ctx, self.ctx.supercells)
        for key, val in get_vasp_force_sets_dict(**forces_dict).items():
            self.ctx[key] = val
            self.out(key, self.ctx[key])

        forces_dict = collect_vasp_forces_and_energies(
            self.ctx, self.ctx.phonon_supercells, prefix="phonon_force_calc"
        )
        for key, val in get_vasp_force_sets_dict(**forces_dict).items():
            self.ctx["phonon_%s" % key] = val
            self.out("phonon_%s" % key, self.ctx["phonon_%s" % key])

    def create_nac_params(self):
        self.report("create nac data")

        calc = self.ctx.born_and_epsilon_calc
        if type(calc) is dict:
            calc_dict = calc
            structure = calc["structure"]
        else:
            calc_dict = calc.outputs
            structure = calc.inputs.structure

        if "born_charges" not in calc_dict:
            raise RuntimeError(
                "Born effective charges could not be found "
                "in the calculation. Please check the calculation setting."
            )
        if "dielectrics" not in calc_dict:
            raise RuntimeError(
                "Dielectric constant could not be found "
                "in the calculation. Please check the calculation setting."
            )

        kwargs = {}
        if self.import_calculations_from_files():
            kwargs["primitive"] = self.ctx.primitive
        self.ctx.nac_params = get_nac_params(
            calc_dict["born_charges"],
            calc_dict["dielectrics"],
            structure,
            self.inputs.symmetry_tolerance,
            **kwargs
        )
        self.out("nac_params", self.ctx.nac_params)

    def run_phono3py_remote(self):
        self.report("remote phonopy calculation")

    def collect_data(self):
        self.report("collect data")

    def create_force_constants(self):
        self.report("create force constants")

    def run_phono3py_in_workchain(self):
        self.report("phonopy calculation in workchain")



"""Utilities related to process builder or inputs dist."""
import copy

from aiida.common import InputValidationError
from aiida.engine import calcfunction
from aiida.orm import Bool, Code, Str, load_group
from aiida.plugins import DataFactory, WorkflowFactory

KpointsData = DataFactory("array.kpoints")
Dict = DataFactory("dict")
StructureData = DataFactory("structure")
PotcarData = DataFactory("vasp.potcar")


def get_workchain_inputs(calculator_inputs, structure, label=None, ctx=None):
    """Return builder inputs of a calculation."""
    if "code" in calculator_inputs.keys():
        code = calculator_inputs["code"]
    else:
        code = Code.get_from_string(calculator_inputs["code_string"])
    plugin_name = code.get_input_plugin_name()
    if plugin_name == "vasp.vasp":
        if isinstance(calculator_inputs["options"], dict):
            options = Dict(dict=calculator_inputs["options"])
        else:
            options = calculator_inputs["options"]
        if isinstance(calculator_inputs["potential_family"], str):
            potential_family = Str(calculator_inputs["potential_family"])
        else:
            potential_family = calculator_inputs["potential_family"]
        if isinstance(calculator_inputs["potential_mapping"], dict):
            potential_mapping = Dict(dict=calculator_inputs["potential_mapping"])
        else:
            potential_mapping = calculator_inputs["potential_mapping"]
        workchain_inputs = {
            "options": options,
            "parameters": _get_parameters_Dict(calculator_inputs),
            "settings": _get_vasp_settings(calculator_inputs),
            "kpoints": _get_kpoints(calculator_inputs, structure),
            "clean_workdir": Bool(False),
            "structure": structure,
            "code": code,
            "potential_family": potential_family,
            "potential_mapping": potential_mapping,
        }
        if label:
            workchain_inputs["metadata"] = {"label": label}
    elif plugin_name == "quantumespresso.pw":
        family = load_group(calculator_inputs["pseudo_family_string"])
        pseudos = family.get_pseudos(structure=structure)
        metadata = {"options": calculator_inputs["options"]}
        if label:
            metadata["label"] = label
        pw = {
            "metadata": metadata,
            "parameters": _get_parameters_Dict(calculator_inputs),
            "structure": structure,
            "pseudos": pseudos,
            "code": code,
        }
        workchain_inputs = {
            "kpoints": _get_kpoints(calculator_inputs, structure),
            "pw": pw,
        }
    elif plugin_name == "quantumespresso.ph":
        qpoints = KpointsData()
        qpoints.set_kpoints_mesh([1, 1, 1], offset=[0, 0, 0])
        metadata = {"options": calculator_inputs["options"]}
        if label:
            metadata["label"] = label
        ph = {
            "metadata": metadata,
            "qpoints": qpoints,
            "parameters": _get_parameters_Dict(calculator_inputs),
            "parent_folder": ctx.nac_params_calcs[0].outputs.remote_folder,
            "code": code,
        }
        workchain_inputs = {"ph": ph}
    else:
        raise RuntimeError("Code could not be found.")

    return workchain_inputs


def get_calculator_process(code_string=None, plugin_name=None):
    """Return WorkChain or CalcJob."""
    if plugin_name is None:
        code = Code.get_from_string(code_string)
        _plugin_name = code.get_input_plugin_name()
    else:
        _plugin_name = plugin_name
    if _plugin_name == "vasp.vasp":
        return WorkflowFactory(_plugin_name)
    elif _plugin_name in ("quantumespresso.pw", "quantumespresso.ph"):
        return WorkflowFactory(_plugin_name + ".base")
    else:
        raise RuntimeError("Code could not be found.")


def get_vasp_immigrant_inputs(folder_path, calculator_settings, label=None):
    """Return VASP immigrant inputs.

    folder_path : str
        VASP directory path.
    calculator_settings : dict
        aiida-phonopy calculator settings for forces or nac params.

    """
    code = Code.get_from_string(calculator_settings["code_string"])

    if code.get_input_plugin_name() == "vasp.vasp":
        inputs = {}
        inputs["code"] = code
        inputs["folder_path"] = Str(folder_path)
        if "settings" in calculator_settings:
            settings = copy.deepcopy(calculator_settings["settings"])
        else:
            settings = {}
        if "parser_settings" in calculator_settings:
            if "parser_settings" in settings:
                settings["parser_settings"].update(
                    calculator_settings["parser_settings"]
                )
            else:
                settings["parser_settings"] = calculator_settings["parser_settings"]
        if settings:
            inputs["settings"] = Dict(dict=settings)
        if "options" in calculator_settings:
            inputs["options"] = Dict(dict=calculator_settings["options"])
        if "metadata" in calculator_settings:
            inputs["metadata"] = calculator_settings["metadata"]
            if label:
                inputs["metadata"]["label"] = label
        elif label:
            inputs["metadata"] = {"label": label}
        if "potential_family" in calculator_settings:
            inputs["potential_family"] = Str(calculator_settings["potential_family"])
        if "potential_mapping" in calculator_settings:
            inputs["potential_mapping"] = Dict(
                dict=calculator_settings["potential_mapping"]
            )
    else:
        raise RuntimeError("Code could not be found.")

    return inputs


def _get_parameters_Dict(calculator_inputs):
    """Return parameters for inputs.parameters.

    If calculator_inputs["parameters"] is already a Dict,
    a new Dict will not be made, and just it will be returned.

    """
    if isinstance(calculator_inputs["parameters"], dict):
        return Dict(dict=calculator_inputs["parameters"])
    else:
        return calculator_inputs["parameters"]


def _get_vasp_settings(calculator_inputs):
    """Update VASP settings.

    If no update of settings and calculator_inputs["settings"] is already a Dict,
    a new Dict will not be made, and just it will be returned.

    """
    updated = False
    if "settings" in calculator_inputs.keys():
        settings = calculator_inputs["settings"]
    else:
        settings = {}
    if "parser_settings" in calculator_inputs.keys():
        settings["parser_settings"] = calculator_inputs["parser_settings"]
        updated = True
    if (
        "parser_settings" not in settings
        or "add_forces" not in settings["parser_settings"]
    ):
        settings["parser_settings"].update({"add_forces": True})
        updated = True

    assert settings

    if updated:
        return create_vasp_inputs_settings(Dict(dict=settings))
    else:
        return settings


@calcfunction
def create_vasp_inputs_settings(settings):
    """Store Dict for VaspWorkChain.inputs.settings."""
    return Dict(dict=settings.get_dict())


def _get_kpoints(calculator_inputs, structure):
    """Return KpointsData."""
    if "kpoints" in calculator_inputs.keys():
        assert isinstance(calculator_inputs["kpoints"], KpointsData)
        return calculator_inputs["kpoints"]
    kpoints = KpointsData()
    kpoints.set_cell_from_structure(structure)
    if "kpoints_density" in calculator_inputs.keys():
        kpoints.set_kpoints_mesh_from_density(calculator_inputs["kpoints_density"])
    elif "kpoints_mesh" in calculator_inputs.keys():
        if "kpoints_offset" in calculator_inputs.keys():
            kpoints_offset = calculator_inputs["kpoints_offset"]
        else:
            kpoints_offset = [0.0, 0.0, 0.0]

        kpoints.set_kpoints_mesh(
            calculator_inputs["kpoints_mesh"], offset=kpoints_offset
        )
    else:
        raise InputValidationError(
            "no kpoint definition in input. "
            "Define either kpoints_density or kpoints_mesh"
        )

    return kpoints



"""Parsers of phonopy output files."""

from aiida.engine import ExitCode
from aiida.common.exceptions import NotExistent
from aiida.parsers.parser import Parser
from aiida.plugins import CalculationFactory
from aiida_phonopy.common.raw_parsers import (
    parse_thermal_properties,
    parse_FORCE_CONSTANTS,
    parse_projected_dos,
    parse_total_dos,
    parse_band_structure,
    parse_phonopy_yaml,
)
from aiida.plugins import DataFactory


Str = DataFactory("str")
PhonopyCalculation = CalculationFactory("phonopy.phonopy")


class PhonopyParser(Parser):
    """Parser the DATA files from phonopy."""

    def __init__(self, calc):
        """Call Parser.__init__."""
        super().__init__(calc)

    def parse(self, **kwargs):
        """Parse retrieved files."""
        self.logger.info("parse retrieved files")

        # select the folder object
        # Check that the retrieved folder is there
        try:
            output_folder = self.retrieved
        except NotExistent:
            return self.exit_codes.ERROR_NO_RETRIEVED_FOLDER

        # check what is inside the folder
        list_of_files = output_folder.list_object_names()

        fc_filename = PhonopyCalculation._INOUT_FORCE_CONSTANTS
        if fc_filename in list_of_files:
            with output_folder.open(fc_filename, "rb") as f:
                self.out("force_constants", parse_FORCE_CONSTANTS(f))

        projected_dos_filename = PhonopyCalculation._OUTPUT_PROJECTED_DOS
        if projected_dos_filename in list_of_files:
            with output_folder.open(projected_dos_filename) as f:
                self.out("pdos", parse_projected_dos(f))

        total_dos_filename = PhonopyCalculation._OUTPUT_TOTAL_DOS
        if total_dos_filename in list_of_files:
            with output_folder.open(total_dos_filename) as f:
                self.out("dos", parse_total_dos(f))

        tp_filename = PhonopyCalculation._OUTPUT_THERMAL_PROPERTIES
        if tp_filename in list_of_files:
            with output_folder.open(tp_filename) as f:
                self.out("thermal_properties", parse_thermal_properties(f))

        band_filename = PhonopyCalculation._OUTPUT_BAND_STRUCTURE
        if band_filename in list_of_files:
            if "symmetry" in self.node.inputs.settings.attributes:
                sym_dataset = self.node.inputs.settings["symmetry"]
                label = "%s (%d)" % (
                    sym_dataset["international"],
                    sym_dataset["number"],
                )
            else:
                label = None
            with output_folder.open(band_filename) as f:
                self.out("band_structure", parse_band_structure(f, label=label))

        options = self.node.get_options()
        phpy_yaml_filename = options["output_filename"]
        if phpy_yaml_filename in list_of_files:
            with output_folder.open(phpy_yaml_filename) as f:
                yaml_dict = parse_phonopy_yaml(f)
                self.out("version", Str(yaml_dict["phonopy"]["version"]))

        self.logger.info("Parsing done.")
        return ExitCode(0)



"""Workflow to calculate NAC params."""
from aiida.engine import WorkChain, append_, calcfunction, if_, while_
from aiida.orm import Code
from aiida.plugins import DataFactory, WorkflowFactory
from phonopy.structure.symmetry import symmetrize_borns_and_epsilon

from aiida_phonopy.common.builders import (
    get_calculator_process,
    get_vasp_immigrant_inputs,
    get_workchain_inputs,
)
from aiida_phonopy.common.utils import (
    get_structure_from_vasp_immigrant,
    phonopy_atoms_from_structure,
)

Float = DataFactory("float")
Str = DataFactory("str")
Dict = DataFactory("dict")
ArrayData = DataFactory("array")
StructureData = DataFactory("structure")


def _get_nac_params(ctx, symmetry_tolerance):
    """Obtain Born effective charges and dielectric constants in primitive cell.

    When Born effective charges and dielectric constants are calculated within
    phonopy workchain, those values are calculated in the primitive cell.
    However using immigrant, the cell may not be primitive cell and can be
    unit cell. In this case, conversion of data is necessary. This conversion
    needs information of the structure where those values were calcualted and
    the target primitive cell structure.

    When using immigrant, structure is in the immigrant calculation but not
    the workchain. 'structure' should be accessible in the vasp immigrant
    workchain level, and this should be fixed in aiida-vasp.

    """
    if ctx.plugin_names[0] == "vasp.vasp":
        calc = ctx.nac_params_calcs[0]
        if "structure" in calc.inputs:
            structure = calc.inputs.structure
        else:
            structure = get_structure_from_vasp_immigrant(calc)
        nac_params = get_vasp_nac_params(
            calc.outputs.born_charges,
            calc.outputs.dielectrics,
            structure,
            symmetry_tolerance,
        )
    elif ctx.plugin_names[0] == "quantumespresso.pw":
        pw_calc = ctx.nac_params_calcs[0]
        ph_calc = ctx.nac_params_calcs[1]
        nac_params = get_qe_nac_params(
            ph_calc.outputs.output_parameters,
            pw_calc.inputs.pw.structure,
            symmetry_tolerance,
        )
    else:
        nac_params = None
    return nac_params


@calcfunction
def get_qe_nac_params(output_parameters, structure, symmetry_tolerance, primitive=None):
    """Return NAC params ArrayData created from QE results."""
    nac_params = _get_nac_params_array(
        output_parameters["effective_charges_eu"],
        output_parameters["dielectric_constant"],
        structure,
        symmetry_tolerance.value,
        primitive=primitive,
    )
    return nac_params


@calcfunction
def get_vasp_nac_params(
    born_charges, epsilon, structure, symmetry_tolerance, primitive=None
):
    """Return NAC params ArrayData created from VASP results."""
    nac_params = _get_nac_params_array(
        born_charges.get_array("born_charges"),
        epsilon.get_array("epsilon"),
        structure,
        symmetry_tolerance.value,
        primitive=primitive,
    )
    return nac_params


def _get_nac_params_array(
    born_charges, epsilon, structure, symmetry_tolerance, primitive=None
):
    phonopy_cell = phonopy_atoms_from_structure(structure)
    if primitive is not None:
        phonopy_primitive = phonopy_atoms_from_structure(primitive)
    else:
        phonopy_primitive = None
    borns_, epsilon_ = symmetrize_borns_and_epsilon(
        born_charges,
        epsilon,
        phonopy_cell,
        symprec=symmetry_tolerance,
        primitive=phonopy_primitive,
    )
    nac_params = ArrayData()
    nac_params.set_array("born_charges", borns_)
    nac_params.set_array("epsilon", epsilon_)
    nac_params.label = "born_charges & epsilon"
    return nac_params


def _get_plugin_names(calculator_settings):
    """Return plugin names of calculators."""
    codes = []
    if "steps" in calculator_settings.keys():
        for step in calculator_settings["steps"]:
            if "code_string" in step.keys():
                code = Code.get_from_string(step["code_string"])
            else:
                code = step["code"]
            codes.append(code)
    else:
        if "code_string" in calculator_settings.keys():
            code = Code.get_from_string(calculator_settings["code_string"])
        else:
            code = calculator_settings["code"]
        codes.append(code)

    plugin_names = []
    for code in codes:
        plugin_names.append(code.get_input_plugin_name())

    return plugin_names


class NacParamsWorkChain(WorkChain):
    """Wrapper to compute non-analytical term correction parameters."""

    @classmethod
    def define(cls, spec):
        """Define inputs, outputs, and outline."""
        super().define(spec)
        spec.input("structure", valid_type=StructureData, required=True)
        spec.input("calculator_inputs", valid_type=dict, required=True, non_db=True)
        spec.input("symmetry_tolerance", valid_type=Float, default=lambda: Float(1e-5))
        spec.input("immigrant_calculation_folder", valid_type=Str, required=False)

        spec.outline(
            cls.initialize,
            while_(cls.continue_calculation)(
                if_(cls.import_calculation_from_files)(
                    cls.read_calculation_from_folder,
                ).else_(
                    cls.run_calculation,
                ),
            ),
            cls.finalize,
        )

        spec.output("nac_params", valid_type=ArrayData, required=True)

        spec.exit_code(
            1001,
            "ERROR_NO_NAC_PARAMS",
            message=("NAC params could not be retrieved from calculaton."),
        )

    def continue_calculation(self):
        """Return boolen for outline."""
        if self.ctx.iteration >= self.ctx.max_iteration:
            return False
        self.ctx.iteration += 1
        return True

    def import_calculation_from_files(self):
        """Return boolen for outline."""
        return "immigrant_calculation_folder" in self.inputs

    def initialize(self):
        """Initialize outline control parameters."""
        self.report("initialization")
        self.ctx.iteration = 0
        if "steps" in self.inputs.calculator_inputs.keys():
            self.ctx.max_iteration = len(self.inputs.calculator_inputs["steps"])
        else:
            self.ctx.max_iteration = 1

        self.ctx.plugin_names = _get_plugin_names(self.inputs.calculator_inputs)

    def run_calculation(self):
        """Run NAC params calculation."""
        self.report(
            "calculation iteration %d/%d" % (self.ctx.iteration, self.ctx.max_iteration)
        )

        if "steps" in self.inputs.calculator_inputs.keys():
            calculator_inputs = self.inputs.calculator_inputs["steps"][
                self.ctx.iteration - 1
            ]
        else:
            calculator_inputs = self.inputs.calculator_inputs
        label = "nac_params_%d" % self.ctx.iteration

        process_inputs = get_workchain_inputs(
            calculator_inputs,
            self.inputs.structure,
            ctx=self.ctx,
            label=label,
        )
        i = self.ctx.iteration - 1
        CalculatorProcess = get_calculator_process(plugin_name=self.ctx.plugin_names[i])
        future = self.submit(CalculatorProcess, **process_inputs)
        self.report("nac_params: {}".format(future.pk))
        self.to_context(nac_params_calcs=append_(future))

    def read_calculation_from_folder(self):
        """Import supercell force calculation using immigrant."""
        self.report(
            "import calculation data in files %d/%d"
            % (self.ctx.iteration, self.ctx.max_iteration)
        )
        label = "nac_params_%d" % self.ctx.iteration
        force_folder = self.inputs.immigrant_calculation_folder
        inputs = get_vasp_immigrant_inputs(
            force_folder.value, self.inputs.calculator_inputs.dict, label=label
        )
        VaspImmigrant = WorkflowFactory("vasp.immigrant")
        future = self.submit(VaspImmigrant, **inputs)
        self.report("nac_params: {}".format(future.pk))
        self.to_context(nac_params_calcs=append_(future))

    def finalize(self):
        """Finalize NAC params calculation."""
        self.report("finalization")

        nac_params = _get_nac_params(self.ctx, self.inputs.symmetry_tolerance)
        if nac_params is None:
            return self.exit_codes.ERROR_NO_NAC_PARAMS

        self.out("nac_params", nac_params)



"""BasePhonopyWorkChain."""

from abc import ABCMeta, abstractmethod

from aiida.engine import WorkChain, if_
from aiida.orm import Code
from aiida.plugins import DataFactory

from aiida_phonopy.common.utils import (
    collect_forces_and_energies,
    get_force_constants,
    get_force_sets,
    get_phonon_properties,
    setup_phonopy_calculation,
)

Float = DataFactory("float")
Bool = DataFactory("bool")
Str = DataFactory("str")
Dict = DataFactory("dict")
ArrayData = DataFactory("array")
XyData = DataFactory("array.xy")
StructureData = DataFactory("structure")
BandsData = DataFactory("array.bands")


class BasePhonopyWorkChain(WorkChain, metaclass=ABCMeta):
    """BasePhonopyWorkchain.

    inputs
    ------
    structure : StructureData
        Unit cell structure.
    phonon_settings : Dict
        Setting to run phonon calculation. Keys are:
        supercell_matrix : list or list of list
            Multiplicity to create supercell from unit cell. Three integer
            values (list) or 3x3 integer values (list of list).
        mesh : list of float, optional
            List of three integer values or float to represent distance between
            neighboring q-points. Default is 100.0.
        distance : float, optional
            Atomic displacement distance. Default is 0.01.
        is_nac : bool, optional
            Deprecated.
            Whether running non-analytical term correction or not. Default is
            False.
        fc_calculator : str
            With this being 'alm', ALM is used to calculate force constants in
            the remote phonopy calculation.
        options : dict
            AiiDA calculation options for phonon calculation used when both of
            run_phonopy and remote_phonopy are True.
    calculator_inputs.force : dict, optional
        This is used for supercell force calculation.
    calculator_inputs.nac : dict, optional
        This is used for Born effective charges and dielectric constant calculation
        in primitive cell. The primitive cell is chosen by phonopy
        automatically.
    subtract_residual_forces : Bool, optional
        Run a perfect supercell force calculation and subtract the residual
        forces from forces in supercells with displacements. Default is False.
    run_phonopy : Bool, optional
        Whether running phonon calculation or not. Default is False.
    remote_phonopy : Bool, optional
        Whether running phonon calculation or not at remote. Default is False.
    code_string : Str, optional
        Code string of phonopy needed when both of run_phonopy and
        remote_phonopy are True.
    symmetry_tolerance : Float, optional
        Symmetry tolerance. Default is 1e-5.

    """

    @classmethod
    def define(cls, spec):
        """Define inputs, outputs, and outline."""
        super().define(spec)
        spec.input("structure", valid_type=StructureData, required=True)
        spec.input("phonon_settings", valid_type=Dict, required=True)
        spec.input_namespace(
            "calculator_inputs", help="Inputs passed to force and NAC calculators."
        )
        spec.input(
            "calculator_inputs.force", valid_type=dict, required=False, non_db=True
        )
        spec.input(
            "calculator_inputs.nac", valid_type=dict, required=False, non_db=True
        )
        spec.input("symmetry_tolerance", valid_type=Float, default=lambda: Float(1e-5))
        spec.input("dry_run", valid_type=Bool, default=lambda: Bool(False))
        spec.input(
            "subtract_residual_forces", valid_type=Bool, default=lambda: Bool(False)
        )
        spec.input("run_phonopy", valid_type=Bool, default=lambda: Bool(False))
        spec.input("remote_phonopy", valid_type=Bool, default=lambda: Bool(False))
        spec.input("displacement_dataset", valid_type=Dict, required=False)
        spec.input("code_string", valid_type=Str, required=False)
        spec.input("code", valid_type=Code, required=False)

        spec.outline(
            cls.initialize,
            cls.run_force_and_nac_calculations,
            if_(cls.dry_run)(cls.postprocess_of_dry_run,).else_(
                cls.create_force_sets,
                if_(cls.is_nac)(cls.attach_nac_params),
                if_(cls.run_phonopy)(
                    if_(cls.remote_phonopy)(
                        cls.run_phonopy_remote,
                        cls.collect_data,
                    ).else_(
                        cls.create_force_constants,
                        cls.run_phonopy_locally,
                    ),
                ),
            ),
            cls.finalize,
        )
        spec.output("force_constants", valid_type=ArrayData, required=False)
        spec.output("primitive", valid_type=StructureData, required=False)
        spec.output("supercell", valid_type=StructureData, required=False)
        spec.output("force_sets", valid_type=ArrayData, required=False)
        spec.output("supercell_forces", valid_type=ArrayData, required=False)
        spec.output("supercell_energy", valid_type=Float, required=False)
        spec.output("nac_params", valid_type=ArrayData, required=False)
        spec.output("thermal_properties", valid_type=XyData, required=False)
        spec.output("band_structure", valid_type=BandsData, required=False)
        spec.output("dos", valid_type=XyData, required=False)
        spec.output("pdos", valid_type=XyData, required=False)
        spec.output("phonon_setting_info", valid_type=Dict)
        spec.exit_code(
            1001,
            "ERROR_NO_PHONOPY_CODE",
            message=(
                "Phonopy Code not found though expected to run phonopy " "remotely."
            ),
        )
        spec.exit_code(
            1002,
            "ERROR_NO_SUPERCELL_MATRIX",
            message=("supercell_matrix was not found."),
        )
        spec.exit_code(
            1003,
            "ERROR_INCONSISTENT_IMMIGRANT_FORCES_FOLDERS",
            message=(
                "Number of supercell folders is different from number "
                "of expected supercells."
            ),
        )

    def dry_run(self):
        """Return boolen for outline."""
        return self.inputs.dry_run

    def remote_phonopy(self):
        """Return boolen for outline."""
        return self.inputs.remote_phonopy

    def run_phonopy(self):
        """Return boolen for outline."""
        return self.inputs.run_phonopy

    def is_nac(self):
        """Return boolen for outline."""
        if "nac" in self.inputs.calculator_inputs:
            return True
        if "is_nac" in self.inputs.phonon_settings.keys():
            self.logger.warning("Use inputs.phonon_settings['is_nac'] is deprecated.")
            return self.inputs.phonon_settings["is_nac"]
        return False

    def initialize(self):
        """Set default settings and create supercells and primitive cell.

        self.ctx.supercells contains supercells as a dict.
        The keys are like 'spercell_000', 'supercell_001', ...,
        where the number of digits depends on the number of supercells.
        'spercell_000' is only available when
            self.inputs.subtract_residual_forces = True.

        """
        self.report("initialization")

        if self.inputs.run_phonopy and self.inputs.remote_phonopy:
            if "code" not in self.inputs and "code_string" not in self.inputs:
                return self.exit_codes.ERROR_NO_PHONOPY_CODE

        if "supercell_matrix" not in self.inputs.phonon_settings.keys():
            return self.exit_codes.ERROR_NO_SUPERCELL_MATRIX

        kwargs = {}
        if "displacement_dataset" in self.inputs:
            kwargs["dataset"] = self.inputs.displacement_dataset
        return_vals = setup_phonopy_calculation(
            self.inputs.phonon_settings,
            self.inputs.structure,
            self.inputs.symmetry_tolerance,
            self.inputs.run_phonopy,
            **kwargs
        )

        for key in ("phonon_setting_info", "primitive", "supercell"):
            self.ctx[key] = return_vals[key]
            self.out(key, self.ctx[key])
        self.ctx.supercells = {}
        if self.inputs.subtract_residual_forces:
            digits = len(str(len(self.ctx.supercells)))
            key = "supercell_%s" % "0".zfill(digits)
            self.ctx.supercells[key] = return_vals["supercell"]
        for key in return_vals:
            if "supercell_" in key:
                self.ctx.supercells[key] = return_vals[key]

    def run_force_and_nac_calculations(self):
        """Run supercell force and NAC params calculations.

        Supercell force calculations and NAC params calculation
        are submitted in this method to make them run in parallel.

        """
        self._run_force_calculations()
        if self.is_nac():
            self._run_nac_params_calculation()

    @abstractmethod
    def _run_force_calculations(self):
        """Run supercell force calculations."""
        pass

    @abstractmethod
    def _run_nac_params_calculation(self):
        """Run nac params calculation."""
        pass

    def postprocess_of_dry_run(self):
        """Show message."""
        self.report("Finish here because of dry-run setting")

    def create_force_sets(self):
        """Build datasets from forces of supercells with displacments."""
        self.report("create force sets")

        forces_dict = collect_forces_and_energies(self.ctx, self.ctx.supercells)
        for key, val in get_force_sets(**forces_dict).items():
            self.ctx[key] = val
            self.out(key, self.ctx[key])

    def attach_nac_params(self):
        """Attach nac_params ArrayData to outputs."""
        self.report("create nac params")

        self.ctx.nac_params = self.ctx.nac_params_calc.outputs.nac_params
        self.out("nac_params", self.ctx.nac_params)

    def run_phonopy_remote(self):
        """Run phonopy at remote computer."""
        self.report("remote phonopy calculation")

        if "code_string" in self.inputs:
            code = Code.get_from_string(self.inputs.code_string.value)
        elif "code" in self.inputs:
            code = self.inputs.code
        builder = code.get_builder()
        builder.structure = self.inputs.structure
        builder.settings = self.ctx.phonon_setting_info
        builder.symmetry_tolerance = self.inputs.symmetry_tolerance
        if "label" in self.inputs.metadata:
            builder.metadata.label = self.inputs.metadata.label
        builder.metadata.options.update(self.inputs.phonon_settings["options"])
        builder.force_sets = self.ctx.force_sets
        if "nac_params" in self.ctx:
            builder.nac_params = self.ctx.nac_params
            builder.primitive = self.ctx.primitive
        future = self.submit(builder)

        self.report("phonopy calculation: {}".format(future.pk))
        self.to_context(**{"phonon_properties": future})
        # return ToContext(phonon_properties=future)

    def collect_data(self):
        """Collect phonon data from remove phonopy calculation."""
        self.report("collect data")
        ph_props = (
            "thermal_properties",
            "dos",
            "pdos",
            "band_structure",
            "force_constants",
        )

        for prop in ph_props:
            if prop in self.ctx.phonon_properties.outputs:
                self.out(prop, self.ctx.phonon_properties.outputs[prop])

        self.report("finish phonon")

    def create_force_constants(self):
        """Create force constants for run_phonopy_locally."""
        self.report("create force constants")

        self.ctx.force_constants = get_force_constants(
            self.inputs.structure,
            self.ctx.phonon_setting_info,
            self.ctx.force_sets,
            self.inputs.symmetry_tolerance,
        )
        self.out("force_constants", self.ctx.force_constants)

    def run_phonopy_locally(self):
        """Run phonopy calculation locally."""
        self.report("phonopy calculation in workchain")

        params = {}
        if "nac_params" in self.ctx:
            params["nac_params"] = self.ctx.nac_params
        result = get_phonon_properties(
            self.inputs.structure,
            self.ctx.phonon_setting_info,
            self.ctx.force_constants,
            self.inputs.symmetry_tolerance,
            **params
        )
        self.out("thermal_properties", result["thermal_properties"])
        self.out("dos", result["dos"])
        self.out("band_structure", result["band_structure"])

        self.report("finish phonon")

    def finalize(self):
        """Show final message."""
        self.report("phonopy calculation has been done.")



from io import StringIO
from phonopy import Phonopy
from phonopy.structure.cells import get_supercell
from phonopy.structure.dataset import forces_in_dataset
from phonopy.interface.phonopy_yaml import PhonopyYaml
from phonopy.file_IO import (
    get_FORCE_SETS_lines,
    get_BORN_lines,
    get_FORCE_CONSTANTS_lines,
)
from aiida_phonopy.common.utils import phonopy_atoms_from_structure


def get_BORN_txt(nac_data, structure, symmetry_tolerance):
    """Returns a string of BORN file.

    nac_data : ArrayData
        Born effective charges and dielectric constants
    structure : StructureData
        This is assumed to be the primitive cell in workchain.
    symmetry_tolerance : float
        Symmetry tolerance.

    """

    born_charges = nac_data.get_array("born_charges")
    epsilon = nac_data.get_array("epsilon")
    pcell = phonopy_atoms_from_structure(structure)
    lines = get_BORN_lines(
        pcell, born_charges, epsilon, symprec=symmetry_tolerance.value
    )

    return "\n".join(lines)


def get_FORCE_SETS_txt(dataset, force_sets=None):
    if dataset is None:
        return None
    if force_sets is None and not forces_in_dataset(dataset):
        return None
    if dataset is not None and force_sets is not None:
        forces = force_sets.get_array("force_sets")
        lines = get_FORCE_SETS_lines(dataset, forces=forces)
    elif dataset is not None and force_sets is None:
        lines = get_FORCE_SETS_lines(dataset)

    return "\n".join(lines)


def get_FORCE_CONSTANTS_txt(force_constants_object):
    force_constants = force_constants_object.get_array("force_constants")
    p2s_map = force_constants_object.get_array("p2s_map")
    lines = get_FORCE_CONSTANTS_lines(force_constants, p2s_map=p2s_map)

    return "\n".join(lines)


def get_phonopy_yaml_txt(
    structure, supercell_matrix=None, primitive_matrix=None, calculator=None
):
    unitcell = phonopy_atoms_from_structure(structure)
    ph = Phonopy(
        unitcell,
        supercell_matrix=supercell_matrix,
        primitive_matrix="auto",
        calculator=calculator,
    )
    phpy_yaml = PhonopyYaml()
    phpy_yaml.set_phonon_info(ph)

    return str(phpy_yaml)


def get_phonopy_options(postprocess_parameters):
    """Return phonopy command option strings."""
    mesh_opts = []
    if "mesh" in postprocess_parameters:
        mesh = postprocess_parameters["mesh"]
        try:
            length = float(mesh)
            mesh_opts.append("--mesh=%f" % length)
        except TypeError:
            mesh_opts.append('--mesh="%d %d %d"' % tuple(mesh))
        mesh_opts.append("--nowritemesh")

    fc_opts = []
    if "fc_calculator" in postprocess_parameters:
        if postprocess_parameters["fc_calculator"].lower().strip() == "alm":
            fc_opts.append("--alm")
    return mesh_opts, fc_opts


def get_disp_fc3_txt(structure, parameters_data, force_sets):
    from phono3py.file_IO import write_cell_yaml

    dataset = force_sets.get_datasets3()
    supercell = get_supercell(
        phonopy_atoms_from_structure(structure),
        parameters_data.dict.supercell,
        symprec=parameters_data.dict.symmetry_tolerance,
    )

    w = StringIO.StringIO()
    w.write("natom: %d\n" % dataset["natom"])

    num_first = len(dataset["first_atoms"])
    w.write("num_first_displacements: %d\n" % num_first)
    if "cutoff_distance" in dataset:
        w.write("cutoff_distance: %f\n" % dataset["cutoff_distance"])

    num_second = 0
    num_disp_files = 0
    for d1 in dataset["first_atoms"]:
        num_disp_files += 1
        num_second += len(d1["second_atoms"])
        for d2 in d1["second_atoms"]:
            if "included" in d2:
                if d2["included"]:
                    num_disp_files += 1
            else:
                num_disp_files += 1

    w.write("num_second_displacements: %d\n" % num_second)
    w.write("num_displacements_created: %d\n" % num_disp_files)

    w.write("first_atoms:\n")
    count1 = 1
    count2 = num_first + 1
    for disp1 in dataset["first_atoms"]:
        disp_cart1 = disp1["displacement"]
        w.write("- number: %5d\n" % (disp1["number"] + 1))
        w.write("  displacement:\n")
        w.write(
            "    [%20.16f,%20.16f,%20.16f ] # %05d\n"
            % (disp_cart1[0], disp_cart1[1], disp_cart1[2], count1)
        )
        w.write("  second_atoms:\n")
        count1 += 1

        included = None
        atom2 = -1
        for disp2 in disp1["second_atoms"]:
            if atom2 != disp2["number"]:
                atom2 = disp2["number"]
                if "included" in disp2:
                    included = disp2["included"]
                pair_distance = disp2["pair_distance"]
                w.write("  - number: %5d\n" % (atom2 + 1))
                w.write("    distance: %f\n" % pair_distance)
                if included is not None:
                    if included:
                        w.write("    included: %s\n" % "true")
                    else:
                        w.write("    included: %s\n" % "false")
                w.write("    displacements:\n")

            disp_cart2 = disp2["displacement"]
            w.write(
                "    - [%20.16f,%20.16f,%20.16f ] # %05d\n"
                % (disp_cart2[0], disp_cart2[1], disp_cart2[2], count2)
            )
            count2 += 1

    write_cell_yaml(w, supercell)
    w.seek(0)
    lines = w.read()
    w.close()
    return lines


def get_forces_txt(force_sets):
    w = StringIO.StringIO()
    from phono3py.file_IO import write_FORCES_FC3

    write_FORCES_FC3(force_sets.get_datasets3(), force_sets.get_forces3(), fp=w)
    w.seek(0)
    lines = w.read()
    w.close()
    return lines


def write_fc2_to_hdf5_file(force_constants, filename):
    from phono3py.file_IO import write_fc2_to_hdf5

    write_fc2_to_hdf5(force_constants.get_data(), filename)


def write_fc3_to_hdf5_file(force_constants, filename):
    from phono3py.file_IO import write_fc3_to_hdf5

    write_fc3_to_hdf5(force_constants.get_data(), filename)


def write_kappa_to_hdf5_file(gp, filename="kappa"):
    import h5py

    with h5py.File(filename, "w") as w:
        for key in gp:
            w.create_dataset(key, data=gp[key])



"""PhonopyWorkChain."""

from aiida.engine import if_, while_
from aiida.plugins import DataFactory

from aiida_phonopy.workflows.forces import ForcesWorkChain
from aiida_phonopy.workflows.nac_params import NacParamsWorkChain
from aiida_phonopy.workflows.phonopy.base import BasePhonopyWorkChain

Dict = DataFactory("dict")
Str = DataFactory("str")


class PhonopyWorkChain(BasePhonopyWorkChain):
    """Phonopy workchain.

    inputs
    ------
    See most of inputs at BasePhonopyWorkChain.
    calculator_settings : Dict
        Deprecated.
        Settings to run force and nac calculations. For example,
            {'forces': force_config,
             'nac': nac_config}
        At least 'forces' key is necessary. 'nac' is optional.
        force_config is used for supercell force calculation. nac_config
        are used for Born effective charges and dielectric constant calculation
        in primitive cell. The primitive cell is chosen by phonopy
        automatically.
    immigrant_calculation_folders : Dict, optional
        'force' key has to exist and 'nac' is necessary when
        phonon_settings['is_nac'] is True. The value of the 'force' key is
        the list of strings of remote directories. The value of 'nac' is the
        string of remote directory.

    """

    @classmethod
    def define(cls, spec):
        """Define inputs, outputs, and outline."""
        super().define(spec)
        spec.input("calculator_settings", valid_type=Dict, required=False)
        spec.input("immigrant_calculation_folders", valid_type=Dict, required=False)

        spec.outline(
            cls.initialize,
            if_(cls.import_calculations)(
                if_(cls.import_calculations_from_files)(
                    cls.initialize_immigrant,
                    while_(cls.continue_import)(
                        cls.read_force_calculations_from_files,
                    ),
                    if_(cls.is_nac)(
                        cls.read_nac_calculations_from_files,
                    ),
                ),
            ).else_(
                cls.run_force_and_nac_calculations,
            ),
            if_(cls.dry_run)(cls.postprocess_of_dry_run,).else_(
                cls.create_force_sets,
                if_(cls.is_nac)(cls.attach_nac_params),
                if_(cls.run_phonopy)(
                    if_(cls.remote_phonopy)(
                        cls.run_phonopy_remote,
                        cls.collect_data,
                    ).else_(
                        cls.create_force_constants,
                        cls.run_phonopy_locally,
                    ),
                ),
            ),
            cls.finalize,
        )

    def import_calculations_from_files(self):
        """Return boolen for outline."""
        return "immigrant_calculation_folders" in self.inputs

    def import_calculations(self):
        """Return boolen for outline."""
        if "immigrant_calculation_folders" in self.inputs:
            return True
        else:
            return False

    def continue_import(self):
        """Return boolen for outline."""
        return self.ctx.num_imported < self.ctx.num_supercell_forces

    def initialize_immigrant(self):
        """Initialize immigrant numbers."""
        self.ctx.num_imported = 0
        self.ctx.num_supercell_forces = len(
            self.inputs.immigrant_calculation_folders["forces"]
        )
        self.ctx.supercell_keys_done = []

        if len(self.ctx.supercells) != self.ctx.num_supercell_forces:
            return self.exit_codes.ERROR_INCONSISTENT_IMMIGRANT_FOLDERS

    def read_force_calculations_from_files(self):
        """Import supercell force calculations.

        Importing backend works only for VASP.

        """
        self.report("import supercell force calculation data in files.")
        num_batch = 50
        self.report("%d calculations per batch." % num_batch)

        calc_folders_Dict = self.inputs.immigrant_calculation_folders
        local_count = 0
        for key, supercell in self.ctx.supercells.items():
            if key in self.ctx.supercell_keys_done:
                continue
            else:
                self.ctx.supercell_keys_done.append(key)
            label = "force_calc_%s" % key.split("_")[-1]
            number = int(key.split("_")[-1])
            if not self.inputs.subtract_residual_forces:
                number -= 1
            force_folder = calc_folders_Dict["forces"][number]
            builder = ForcesWorkChain.get_builder()
            builder.metadata.label = label
            builder.structure = supercell
            calculator_settings = self.inputs.calculator_settings
            builder.calculator_inputs = calculator_settings
            builder.immigrant_calculation_folder = Str(force_folder)
            future = self.submit(builder)
            self.report("{} pk = {}".format(label, future.pk))
            self.to_context(**{label: future})
            self.ctx.num_imported += 1
            if not self.continue_import():
                break
            local_count += 1
            if local_count == num_batch:
                break

    def read_nac_calculations_from_files(self):
        """Import NAC params calculation.

        Importing backend works only for VASP.

        """
        self.report("import NAC calculation data in files")
        label = "nac_params_calc"
        calc_folders_Dict = self.inputs.immigrant_calculation_folders
        nac_folder = calc_folders_Dict["nac"][0]
        builder = NacParamsWorkChain.get_builder()
        builder.metadata.label = label
        builder.structure = self.ctx.primitive
        calculator_settings = self.inputs.calculator_settings
        builder.calculator_inputs = calculator_settings
        builder.immigrant_calculation_folder = Str(nac_folder)
        future = self.submit(builder)
        self.report("{} pk = {}".format(label, future.pk))
        self.to_context(**{label: future})

    def _run_force_calculations(self):
        """Run supercell force calculations."""
        self.report("run force calculations")

        for key, supercell in self.ctx.supercells.items():
            label = "force_calc_%s" % key.split("_")[-1]
            builder = ForcesWorkChain.get_builder()
            builder.metadata.label = label
            builder.structure = supercell
            if "force" in self.inputs.calculator_inputs:
                calculator_inputs = self.inputs.calculator_inputs.force
            else:
                calculator_inputs = self.inputs.calculator_settings["forces"]
                self.logger.warning(
                    "Use calculator_inputs.force instead of "
                    "calculator_settings['forces']."
                )
            builder.calculator_inputs = calculator_inputs
            future = self.submit(builder)
            self.report("{} pk = {}".format(label, future.pk))
            self.to_context(**{label: future})

    def _run_nac_params_calculation(self):
        """Run nac params calculation."""
        self.report("run nac params calculation")

        builder = NacParamsWorkChain.get_builder()
        builder.metadata.label = "nac_params"
        builder.structure = self.ctx.primitive
        if "nac" in self.inputs.calculator_inputs:
            calculator_inputs = self.inputs.calculator_inputs.nac
        else:
            calculator_inputs = self.inputs.calculator_settings["nac"]
            self.logger.warning(
                "Use calculator_inputs.nac instead of calculator_settings['nac']."
            )

        builder.calculator_inputs = calculator_inputs
        future = self.submit(builder)
        self.report("nac_params: {}".format(future.pk))
        self.to_context(**{"nac_params_calc": future})



"""File parsers."""

import h5py
import yaml

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader
import numpy as np
from aiida.plugins import DataFactory
from aiida_phonopy.common.utils import (
    get_total_dos,
    get_projected_dos,
    get_thermal_properties,
    get_bands,
)


ArrayData = DataFactory("array")


def parse_phonopy_yaml(f):
    """phonopy.yaml parser."""
    data = yaml.load(f, Loader=Loader)
    return data


def parse_FORCE_CONSTANTS(f):
    """force_constants.hdf5 parser."""
    with h5py.File(f, "r") as f5:
        force_constants = f5["force_constants"][:]
    fc_array = ArrayData()
    fc_array.set_array("force_constants", force_constants)
    fc_array.label = "force_constants"
    return fc_array


def parse_total_dos(f):
    """total_dos.dat parser."""
    data = np.loadtxt(f)
    total_dos = {"frequency_points": data[:, 0], "total_dos": data[:, 1]}
    dos = get_total_dos(total_dos)
    return dos


def parse_projected_dos(f):
    """projected_dos.dat parser."""
    data = np.loadtxt(f)
    projected_dos = {"frequency_points": data[:, 0], "projected_dos": data[:, 1:].T}
    pdos = get_projected_dos(projected_dos)
    return pdos


def parse_thermal_properties(f):
    """thermal_properties.yaml parser."""
    thermal_properties = {
        "temperatures": [],
        "free_energy": [],
        "entropy": [],
        "heat_capacity": [],
    }
    data = yaml.load(f, Loader=Loader)
    for tp in data["thermal_properties"]:
        thermal_properties["temperatures"].append(tp["temperature"])
        thermal_properties["entropy"].append(tp["entropy"])
        thermal_properties["free_energy"].append(tp["free_energy"])
        thermal_properties["heat_capacity"].append(tp["heat_capacity"])
    for key in thermal_properties:
        thermal_properties[key] = np.array(thermal_properties[key])

    tprops = get_thermal_properties(thermal_properties)

    return tprops


def parse_band_structure(f, label=None):
    """band.yaml parser."""
    bands = yaml.load(f, Loader=Loader)
    frequencies_flat = []
    qpoints_flat = []
    for k in bands["phonon"]:
        frequencies_flat.append([b["frequency"] for b in k["band"]])
        qpoints_flat.append(k["q-position"])

    frequencies = []
    qpoints = []
    for n in bands["segment_nqpoint"]:
        qpoints.append([qpoints_flat.pop(0) for i in range(n)])
        frequencies.append([frequencies_flat.pop(0) for i in range(n)])

    label_pairs = []
    for pair in bands["labels"]:
        label_pairs.append(
            [
                x.replace("$", "")
                .replace("\\", "")
                .replace("mathrm{", "")
                .replace("}", "")
                .upper()
                for x in pair
            ]
        )

    labels = [label_pairs[0][0], label_pairs[0][1]]
    path_connections = []
    for i, pairs in enumerate(label_pairs[1:]):
        if pairs[0] == label_pairs[i][1]:
            labels.append(pairs[1])
            path_connections.append(True)
        else:
            labels += pairs
            path_connections.append(False)
    path_connections.append(False)

    bs = get_bands(qpoints, frequencies, labels, path_connections, label=label)

    return bs


def parse_kappa(f):
    """kappa.hdf5 parser."""
    kappa = ArrayData()
    f = h5py.File(f, "r")
    for item in f:
        array = np.array(f[item])
        kappa.set_array(item, array)
    f.close()

    return kappa



"""General utilities."""

import numpy as np
from aiida.engine import calcfunction
from aiida.orm import Bool, Float, Int, Str, load_node
from aiida.plugins import DataFactory
from phonopy import Phonopy
from phonopy.interface.calculator import (
    get_default_displacement_distance,
    get_default_physical_units,
)
from phonopy.structure.atoms import PhonopyAtoms

Dict = DataFactory("dict")
ArrayData = DataFactory("array")
XyData = DataFactory("array.xy")
StructureData = DataFactory("structure")
BandsData = DataFactory("array.bands")


@calcfunction
def get_remote_fc_calculation_settings(phonon_settings):
    """Create remote force constants phonopy calculation setting.

    keys condidered:
        supercell_matrix
        fc_calculator

    """
    key = "supercell_matrix"
    if key in phonon_settings.dict:
        fc_settings = {key: phonon_settings[key]}
    else:
        return None
    key = "fc_calculator"
    if key in phonon_settings.dict:
        fc_settings["postprocess_parameters"] = {key: phonon_settings[key]}
    return Dict(dict=fc_settings)


@calcfunction
def setup_phonopy_calculation(
    phonon_settings, structure, symmetry_tolerance, run_phonopy, dataset=None
):
    """Set up phonopy calculation.

    Valid keys in phonon_settings_info are
        ('supercell_matrix',
         'phonon_supercell_matrix',
         'distance',
         'symmetry_tolerance',
         'number_of_snapshots',
         'random_seed',
         'is_plusminus',
         'is_diagonal',
         'is_trigonal',
         'mesh',
         'fc_calculator').

    Returns
    -------
    dict
        'supercell' : StructureData
            Perfect supercell.
        'supercell_001', 'supercell_002', ... : StructureData
            Supercells with displacements
        'primitive' : StructureData
            Primitive cell.
        'phonon_supercell' : StructureData
            For phono3py. Perfect supercell for harmonic phonon calculation.
        'phonon_supercell_001', 'phonon_supercell_002', ... : StructureData
            For phono3py. Supercells with displacements for harmonic phonon
            calculation.
        'phonon_setting_info' : Dict
            Phonopy setting parameters including those generated in the
            process of displacements creation, e.g., primitive and  sueprcell
            matrix and symmetry information.

    phonon_setting_info contains the following entries:
        'version' : str
            Phonopy version number.
        'supercell_matrix' : array_like
            3x3 integer matrix to generate supercell matrix.
        'phonon_supercell_matrix' : array_like
            3x3 integer matrix to generate fc2 supercell matrix for Phono3py.
        'distance' : float
            Displacement distance.
        'symmetry_tolerance' : float
            Tolerance length used for symmetry finding.
        'displacement_dataset' : dict
            Phonopy.dataset or Phono3py.dataset.
        'primitive_matrix' : array_like
            Phonopy.primitive_matrix.
        'symmetry' : dict
            'number' : Space group number.
            'international' : Space group type.
        'phonon_displacement_dataset' : dict
            Phono3py.phonon_dataset.
        'number_of_snapshots' : int
        'random_seed' : int
        'is_plusminus' : bool
        'is_diagonal' : bool
        'is_trigonal' : bool
        'postprocess_parameters' : dict
            This is given when run_phonopy=True.
            'mesh' : float or list
                Mesh numbers or distance measure of q-point sampling mesh.
            'fc_calculator' : str
                External force constants calculator.

    """
    ph_settings = _get_setting_info(phonon_settings)
    if run_phonopy:
        params = _get_phonopy_postprocess_info(phonon_settings)
        ph_settings["postprocess_parameters"] = params

    ph = _get_phonopy_instance(
        structure, ph_settings, symmetry_tolerance=symmetry_tolerance.value
    )
    ph_settings["version"] = ph.version

    if dataset is None:
        supported_keys = (
            "distance",
            "is_plusminus",
            "is_diagonal",
            "is_trigonal",
            "number_of_snapshots",
            "random_seed",
        )
        kwargs = {key: ph_settings[key] for key in ph_settings if key in supported_keys}
        ph.generate_displacements(**kwargs)
    else:
        ph.dataset = dataset.get_dict()

    _update_structure_info(ph_settings, ph)
    structures_dict = _generate_phonon_structures(ph)
    return_vals = {"phonon_setting_info": Dict(dict=ph_settings)}
    return_vals.update(structures_dict)

    return return_vals


def _get_phonopy_postprocess_info(phonon_settings):
    """Return phonopy postprocess parameters."""
    valid_keys = ("mesh", "fc_calculator")
    params = {}
    for key in valid_keys:
        if key in phonon_settings.keys():
            params[key] = phonon_settings[key]

    if "mesh" not in phonon_settings.keys():
        params["mesh"] = 100.0
    return params


@calcfunction
def generate_phono3py_cells(
    phonon_settings, structure, symmetry_tolerance, dataset=None
):
    """Set up phono3py calculation.

    Returns
    -------
    dict
        'supercell' : StructureData
            Perfect supercell.
        'supercell_001', 'supercell_002', ... : StructureData
            Supercells with displacements
        'primitive' : StructureData
            Primitive cell.
        'phonon_supercell' : StructureData
            For phono3py. Perfect supercell for harmonic phonon calculation.
        'phonon_supercell_001', 'phonon_supercell_002', ... : StructureData
            For phono3py. Supercells with displacements for harmonic phonon
            calculation.
        'phonon_setting_info' : Dict
            Phonopy setting parameters including those generated in the
            process of displacements creation, e.g., primitive and  sueprcell
            matrix and symmetry information.

    phonon_setting_info contains the following entries:
        'supercell_matrix' : array_like
            3x3 integer matrix to generate supercell matrix.
        'phonon_supercell_matrix' : array_like
            3x3 integer matrix to generate fc2 supercell matrix for Phono3py.
        'distance' : float
            Displacement distance.
        'symmetry_tolerance' : float
            Tolerance length used for symmetry finding.
        'displacement_dataset' : dict
            Phonopy.dataset or Phono3py.dataset.
        'primitive_matrix' : array_like
            Phonopy.primitive_matrix.
        'symmetry' : dict
            'number' : Space group number.
            'international' : Space group type.
        'phonon_displacement_dataset' : dict
            Phono3py.phonon_dataset.
        'is_plusminus' : bool
        'is_diagonal' : bool

    """
    ph_settings = _get_setting_info(phonon_settings, code_name="phono3py")

    ph = _get_phono3py_instance(structure, ph_settings, {})
    if dataset is None:
        ph.generate_displacements(distance=ph_settings["distance"])
    else:
        ph.dataset = dataset.get_dict()

    _update_structure_info(ph_settings, ph)
    structures_dict = _generate_phonon_structures(ph)
    return_vals = {"phonon_setting_info": Dict(dict=ph_settings)}
    return_vals.update(structures_dict)

    return return_vals


@calcfunction
def get_force_constants(structure, phonon_settings, force_sets, symmetry_tolerance):
    """Calculate force constants."""
    phonon = _get_phonopy_instance(
        structure, phonon_settings, symmetry_tolerance=symmetry_tolerance.value
    )
    phonon.dataset = phonon_settings["displacement_dataset"]
    phonon.forces = force_sets.get_array("force_sets")

    if "fc_calculator" in phonon_settings["postprocess_parameters"]:
        postprocess_parameters = phonon_settings["postprocess_parameters"]
        if postprocess_parameters["fc_calculator"].lower().strip() == "alm":
            phonon.produce_force_constants(fc_calculator="alm")
    else:
        phonon.produce_force_constants()
    force_constants = ArrayData()
    force_constants.set_array("force_constants", phonon.force_constants)
    force_constants.set_array("p2s_map", phonon.primitive.p2s_map)
    force_constants.label = "force_constants"

    return force_constants


@calcfunction
def get_phonon_properties(
    structure, phonon_settings, force_constants, symmetry_tolerance, nac_params=None
):
    """Calculate phonon properties."""
    phonon_settings_dict = phonon_settings.get_dict()
    ph = _get_phonopy_instance(
        structure,
        phonon_settings_dict,
        symmetry_tolerance=symmetry_tolerance.value,
        nac_params=nac_params,
    )
    ph.force_constants = force_constants.get_array("force_constants")
    mesh = phonon_settings["postprocess_parameters"]["mesh"]

    # Mesh
    total_dos, pdos, thermal_properties = get_mesh_property_data(ph, mesh)

    # Band structure
    bs = _get_bands_data(ph)

    return {
        "dos": total_dos,
        "pdos": pdos,
        "thermal_properties": thermal_properties,
        "band_structure": bs,
    }


@calcfunction
def get_data_from_node_id(node_id):
    """Collect VASP output from node_id."""
    n = load_node(node_id.value)
    if "structure" in n.inputs:
        cell = phonopy_atoms_from_structure(n.inputs.structure)
        structure = phonopy_atoms_to_structure(cell)
    else:
        raise RuntimeError("Crystal structure could not be found.")

    if "born_charges" in n.outputs and "dielectrics" in n.outputs:
        born = ArrayData()
        born.set_array("born_charges", n.outputs.born_charges.get_array("born_charges"))
        born.label = "born_charges"
        epsilon = ArrayData()
        epsilon.set_array("epsilon", n.outputs.dielectrics.get_array("epsilon"))
        epsilon.label = "epsilon"
        return {"born_charges": born, "dielectrics": epsilon, "structure": structure}
    elif "forces" in n.outputs:
        forces = ArrayData()
        forces.set_array("final", n.outputs.forces.get_array("final"))
        forces.label = "forces"
        return {"forces": forces, "structure": structure}
    else:
        raise RuntimeError("Forces or NAC params were not found.")


def compare_structures(cell_ref, cell_calc, symmetry_tolerance):
    """Compare two PhonopyAtoms instances."""
    symprec = symmetry_tolerance.value
    cell_diff = np.subtract(cell_ref.cell, cell_calc.cell)
    if (np.abs(cell_diff) > symprec).any():
        succeeded = Bool(False)
        succeeded.label = "False"
        return succeeded

    positions_ref = [site.position for site in cell_ref.sites]
    positions_calc = [site.position for site in cell_calc.sites]
    diff = np.subtract(positions_ref, positions_calc)
    diff -= np.rint(diff)
    dist = np.sqrt(np.sum(np.dot(diff, cell_ref.cell) ** 2, axis=1))
    if (dist > symprec).any():
        succeeded = Bool(False)
        succeeded.label = "False"
        return succeeded

    succeeded = Bool(True)
    succeeded.label = "True"
    return succeeded


def get_structure_from_vasp_immigrant(calc_node):
    """Get structure from VASP immigrant workchain."""
    for lt in calc_node.get_outgoing():
        if "iteration_" in lt.link_label:
            structure = lt.node.inputs.structure
            return structure
    return None


def get_mesh_property_data(ph, mesh):
    """Return total DOS, PDOS, thermal properties."""
    ph.set_mesh(mesh)
    ph.run_total_dos()

    dos = get_total_dos(ph.get_total_dos_dict())

    ph.run_thermal_properties()
    tprops = get_thermal_properties(ph.get_thermal_properties_dict())

    ph.set_mesh(mesh, is_eigenvectors=True, is_mesh_symmetry=False)
    ph.run_projected_dos()
    pdos = get_projected_dos(ph.get_projected_dos_dict())

    return dos, pdos, tprops


def get_total_dos(total_dos):
    """Return XyData of total DOS."""
    dos = XyData()
    dos.set_x(total_dos["frequency_points"], "Frequency", "THz")
    dos.set_y(total_dos["total_dos"], "Total DOS", "1/THz")
    dos.label = "Total DOS"
    return dos


def get_projected_dos(projected_dos):
    """Return XyData of PDOS."""
    pdos = XyData()
    pdos_list = [pd for pd in projected_dos["projected_dos"]]
    pdos.set_x(projected_dos["frequency_points"], "Frequency", "THz")
    pdos.set_y(
        pdos_list,
        [
            "Projected DOS",
        ]
        * len(pdos_list),
        [
            "1/THz",
        ]
        * len(pdos_list),
    )
    pdos.label = "Projected DOS"
    return pdos


def get_thermal_properties(thermal_properties):
    """Return XyData of thermal properties."""
    tprops = XyData()
    tprops.set_x(thermal_properties["temperatures"], "Temperature", "K")
    tprops.set_y(
        [
            thermal_properties["free_energy"],
            thermal_properties["entropy"],
            thermal_properties["heat_capacity"],
        ],
        ["Helmholtz free energy", "Entropy", "Cv"],
        ["kJ/mol", "J/K/mol", "J/K/mol"],
    )
    tprops.label = "Thermal properties"
    return tprops


def _get_bands_data(ph):
    ph.auto_band_structure()
    labels = [
        x.replace("$", "")
        .replace("\\", "")
        .replace("mathrm{", "")
        .replace("}", "")
        .upper()
        for x in ph.band_structure.labels
    ]
    frequencies = ph.band_structure.frequencies
    qpoints = ph.band_structure.qpoints
    path_connections = ph.band_structure.path_connections
    label = "%s (%d)" % (
        ph.symmetry.dataset["international"],
        ph.symmetry.dataset["number"],
    )

    return get_bands(qpoints, frequencies, labels, path_connections, label=label)


def get_bands(qpoints, frequencies, labels, path_connections, label=None):
    """Return BandsData."""
    qpoints_list = list(qpoints[0])
    frequencies_list = list(frequencies[0])
    labels_list = [
        (0, labels[0]),
    ]
    label_index = 1

    for pc, qs, fs in zip(path_connections[:-1], qpoints[1:], frequencies[1:]):
        if labels[label_index] == "GAMMA" and pc:
            labels_list.append((len(qpoints_list) - 1, labels[label_index]))
            if label_index < len(labels):
                labels_list.append((len(qpoints_list), labels[label_index]))
            label_index += 1
            qpoints_list += list(qs)
            frequencies_list += list(fs)
        elif pc:
            labels_list.append((len(qpoints_list) - 1, labels[label_index]))
            label_index += 1
            qpoints_list += list(qs[1:])
            frequencies_list += list(fs[1:])
        else:
            labels_list.append((len(qpoints_list) - 1, labels[label_index]))
            label_index += 1
            if label_index < len(labels):
                labels_list.append((len(qpoints_list), labels[label_index]))
                label_index += 1
            qpoints_list += list(qs)
            frequencies_list += list(fs)
    labels_list.append((len(qpoints_list) - 1, labels[-1]))

    bs = BandsData()
    bs.set_kpoints(np.array(qpoints_list))
    bs.set_bands(np.array(frequencies_list), units="THz")
    bs.labels = labels_list
    if label is not None:
        bs.label = label

    return bs


def _get_phonopy_instance(
    structure, phonon_settings_dict, symmetry_tolerance=1e-5, nac_params=None
):
    """Create Phonopy instance."""
    phonon = Phonopy(
        phonopy_atoms_from_structure(structure),
        supercell_matrix=phonon_settings_dict["supercell_matrix"],
        primitive_matrix="auto",
        symprec=symmetry_tolerance,
    )
    if nac_params:
        units = get_default_physical_units("vasp")
        factor = units["nac_factor"]
        _nac_params = {
            "born": nac_params.get_array("born_charges"),
            "dielectric": nac_params.get_array("epsilon"),
            "factor": factor,
        }
        phonon.nac_params = _nac_params

    return phonon


def _get_phono3py_instance(structure, phonon_settings_dict, params):
    """Create Phono3py instance."""
    from phono3py import Phono3py

    if "phonon_supercell_matrix" in phonon_settings_dict:
        ph_smat = phonon_settings_dict["phonon_supercell_matrix"]
    else:
        ph_smat = None
    ph3py = Phono3py(
        phonopy_atoms_from_structure(structure),
        supercell_matrix=phonon_settings_dict["supercell_matrix"],
        primitive_matrix="auto",
        phonon_supercell_matrix=ph_smat,
        symprec=phonon_settings_dict["symmetry_tolerance"],
    )
    if "nac_params" in params:
        from phonopy.interface.calculator import get_default_physical_units

        units = get_default_physical_units("vasp")
        factor = units["nac_factor"]
        nac_params = {
            "born": params["nac_params"].get_array("born_charges"),
            "dielectric": params["nac_params"].get_array("epsilon"),
            "factor": factor,
        }
        ph3py.nac_params = nac_params

    return ph3py


def phonopy_atoms_to_structure(cell):
    """Convert PhonopyAtoms to StructureData."""
    symbols = cell.symbols
    positions = cell.positions
    structure = StructureData(cell=cell.cell)
    for symbol, position in zip(symbols, positions):
        structure.append_atom(position=position, symbols=symbol)
    return structure


def phonopy_atoms_from_structure(structure):
    """Convert StructureData to PhonopyAtoms."""
    cell = PhonopyAtoms(
        symbols=[site.kind_name for site in structure.sites],
        positions=[site.position for site in structure.sites],
        cell=structure.cell,
    )
    return cell


def from_node_id_to_aiida_node_id(node_id):
    """Convert PK or UUID to an AiiDA data type."""
    if type(node_id) is int:
        return Int(node_id)
    elif type(node_id) is str:
        return Str(node_id)
    else:
        raise RuntimeError("%s is not supported in load_node." % type(node_id))


def collect_forces_and_energies(ctx, ctx_supercells, prefix="force_calc"):
    """Collect forces and energies from calculation outputs.

    Parameters
    ----------
    ctx : AttributeDict-like
        AiiDA workchain context.
    ctx_supercells : dict of StructDict
        Supercells. For phono3py, this can be phonon_supercells.
    prefix : str
        Prefix string of dictionary keys of ctx.

    Returns
    -------
    dict
        Forces and energies.

    """
    forces_dict = {}
    for key in ctx_supercells:
        # key: e.g. "supercell_001", "phonon_supercell_001"
        num = key.split("_")[-1]  # e.g. "001"
        calc = ctx["%s_%s" % (prefix, num)]
        if type(calc) is dict:
            calc_dict = calc
        else:
            calc_dict = calc.outputs
        forces_dict["forces_%s" % num] = calc_dict["forces"]
        if "energy" in calc_dict:
            forces_dict["energy_%s" % num] = calc_dict["energy"]
        elif "total_energy" in calc_dict:  # For CommonWorkflow
            forces_dict["energy_%s" % num] = calc_dict["total_energy"]

    return forces_dict


@calcfunction
def get_force_sets(**forces_dict):
    """Create force sets from supercell forces.

    Parameters
    ----------
    forces_dict : dict
        'forces_001', 'forces_002', ... have to exist.
        'energy_001', 'energy_002', ... are optional.
        'forces_000' and 'energy_000' for perfect supercell are optional.
        The zero-padding length of the numbers can change depending on total
        number of supercell calculations.

    """
    (force_sets, energies, forces_0_key, energy_0_key) = _get_force_set(**forces_dict)

    force_sets_data = ArrayData()
    force_sets_data.set_array("force_sets", force_sets)
    if energies is not None:
        force_sets_data.set_array("energies", energies)
    force_sets_data.label = "force_sets"
    ret_dict = {"force_sets": force_sets_data}
    if forces_0_key is not None:
        ret_dict["supercell_forces"] = forces_dict[forces_0_key]
    if energy_0_key is not None:
        ret_dict["supercell_energy"] = forces_dict[energy_0_key]

    return ret_dict


def _get_force_set(**forces_dict):
    num_forces = 0
    num_energies = 0
    forces_0_key = None
    energy_0_key = None
    shape = None
    for key in forces_dict:
        value = forces_dict[key]
        if int(key.split("_")[-1]) != 0:
            if "forces" in key:
                num_forces += 1
                if shape is None:
                    shape = value.get_array("forces").shape
            elif "energy" in key:
                num_energies += 1
        else:
            if "forces" in key:
                forces_0_key = key
            elif "energy" in key:
                energy_0_key = key

    force_sets = np.zeros((num_forces,) + shape, dtype=float)
    if num_energies > 0:
        energies = np.zeros(num_energies, dtype=float)
    else:
        energies = None
    if forces_0_key is None:
        forces_0 = None
    else:
        forces_0 = forces_dict[forces_0_key].get_array("forces")

    for key in forces_dict:
        value = forces_dict[key]
        num = int(key.split("_")[-1])  # e.g. "001" --> 1
        if "forces" in key:
            forces = value.get_array("forces")
            if forces_0 is None:
                force_sets[num - 1] = forces
            else:
                force_sets[num - 1] = forces - forces_0
        elif "energy" in key:
            if isinstance(value, Float):  # For CommonWorkflow
                energies[num - 1] = value.value
            else:
                energies[num - 1] = value.get_array("energy")[-1]

    return force_sets, energies, forces_0_key, energy_0_key


def collect_vasp_forces_and_energies(ctx, ctx_supercells, prefix="force_calc"):
    """Collect forces and energies from calculation outputs.

    Parameters
    ----------
    ctx : AttributeDict-like
        AiiDA workchain context.
    ctx_supercells : dict of StructDict
        Supercells. For phono3py, this can be phonon_supercells.
    prefix : str
        Prefix string of dictionary keys of ctx.

    Returns
    -------
    dict
        Forces and energies.

    """
    forces_dict = {}
    for key in ctx_supercells:
        # key: e.g. "supercell_001", "phonon_supercell_001"
        num = key.split("_")[-1]  # e.g. "001"
        calc = ctx["%s_%s" % (prefix, num)]
        if type(calc) is dict:
            calc_dict = calc
        else:
            calc_dict = calc.outputs
        if "forces" in calc_dict and "final" in calc_dict["forces"].get_arraynames():
            forces_dict["forces_%s" % num] = calc_dict["forces"]
        else:
            raise RuntimeError("Forces could not be found in calculation %s." % num)

        if (
            "misc" in calc_dict and "total_energies" in calc_dict["misc"].keys()
        ):  # needs .keys()
            forces_dict["misc_%s" % num] = calc_dict["misc"]

    return forces_dict


@calcfunction
def get_vasp_force_sets_dict(**forces_dict):
    """Create force sets from supercell forces."""
    forces = []
    energies = []
    forces_0 = None
    energy_0 = None

    for key in forces_dict:
        num = int(key.split("_")[-1])
        if num == 0:
            continue
        if "forces" in key:
            forces.append(None)
        elif "misc" in key:
            energies.append(None)

    for key in forces_dict:
        num = int(key.split("_")[-1])  # e.g. "001" --> 1
        if "forces" in key:
            forces_ndarray = forces_dict[key].get_array("final")
            if num == 0:
                forces_0 = forces_ndarray
            else:
                forces[num - 1] = forces_ndarray
        elif "misc" in key:
            for energy_key in ("energy_extrapolated",):
                if energy_key in forces_dict[key]["total_energies"]:
                    energy = forces_dict[key]["total_energies"][energy_key]
                if num == 0:
                    energy_0 = energy
                else:
                    energies[num - 1] = energy
                break

    if forces_0 is not None:
        for forces_ndarray in forces:
            forces_ndarray -= forces_0

    force_sets = ArrayData()
    force_sets.set_array("force_sets", np.array(forces))
    if energies:
        force_sets.set_array("energies", np.array(energies))
    force_sets.label = "force_sets"
    ret_dict = {"force_sets": force_sets}
    if forces_0 is not None:
        forces_0_array = ArrayData()
        forces_0_array.set_array("forces", forces_0)
        ret_dict["supercell_forces"] = forces_0_array
    if energy_0 is not None:
        ret_dict["supercell_energy"] = Float(energy_0)

    return ret_dict


def _get_setting_info(phonon_settings, code_name="phonopy"):
    """Convert AiiDA inputs to a dict.

    code_name : 'phonopy' or 'phono3py'

    Note
    ----
    Designed to be shared by phonopy and phono3py.

    Returns
    -------
    dict
        'supercell_matrix' : ndarray
            3x3 integer matrix to generate supercell matrix.
        'phonon_supercell_matrix' : ndarray
            3x3 integer matrix to generate fc2 supercell matrix for Phono3py.
        'mesh' : list or float
            Mesh numbers or a distance to represent mesh numbers.
        'distance' : float
            Displacement distance.
        'symmetry_tolerance' : float
            Tolerance length used for symmetry finding.

    """
    ph_settings = {}
    valid_keys = (
        "supercell_matrix",
        "phonon_supercell_matrix",
        "distance",
        "symmetry_tolerance",
        "number_of_snapshots",
        "random_seed",
        "is_plusminus",
        "is_diagonal",
        "is_trigonal",
    )
    for key, value in phonon_settings.get_dict().items():
        if key in valid_keys:
            ph_settings[key] = value
    dim = ph_settings["supercell_matrix"]
    ph_settings["supercell_matrix"] = _get_supercell_matrix(dim)
    if "phonon_supercell_matrix" in ph_settings:
        dim_fc2 = ph_settings["phonon_supercell_matrix"]
        ph_settings["phonon_supercell_matrix"] = _get_supercell_matrix(
            dim_fc2, smat_type="phonon_supercell_matrix"
        )
    if "distance" not in ph_settings:
        if code_name == "phono3py":
            from phono3py.interface.calculator import (
                get_default_displacement_distance as get_phono3py_ddd,
            )

            distance = get_phono3py_ddd("vasp")
        else:
            distance = get_default_displacement_distance("vasp")
        ph_settings["distance"] = distance

    return ph_settings


def _get_supercell_matrix(dim, smat_type="supercell_matrix"):
    if len(np.ravel(dim)) == 3:
        smat = np.diag(dim)
    else:
        smat = np.array(dim)
    if not np.issubdtype(smat.dtype, np.integer):
        raise TypeError("%s is not integer matrix." % smat_type)
    else:
        return smat.tolist()


def _update_structure_info(ph_settings, ph):
    """Update a phonon_settings dict.

    Parameters
    ----------
    ph_settings : Dict
         Phonopy setting information.
    ph : Phonopy or Phono3py
         A Phonopy or Phono3py instance.

    Returns
    -------
    dict
        'displacement_dataset' : dict
            Phonopy.dataset or Phono3py.dataset.
        'primitive_matrix' : ndarray
            Phonopy.primitive_matrix.
        'symmetry' : dict
            'number' : Space group number.
            'international' : Space group type.
        'phonon_displacement_dataset' : dict
            Phono3py.phonon_dataset.

    """
    ph_settings["displacement_dataset"] = ph.dataset
    ph_settings["primitive_matrix"] = ph.primitive_matrix
    ph_settings["symmetry"] = {
        "number": ph.symmetry.dataset["number"],
        "international": ph.symmetry.dataset["international"],
    }

    if "phonon_supercell_matrix" in ph.__dir__():
        if ph.phonon_supercell_matrix is not None:
            ph_settings["phonon_displacement_dataset"] = ph.phonon_dataset

    return ph_settings


def _generate_phonon_structures(ph):
    """Generate AiiDA structures of phonon related cells.

    Note
    ----
    Designed to be shared by phonopy and phono3py.
    ph is either an instance of Phonopy or Phono3py.

    Returns
    -------
    dict of StructureData
        'supercell'
            Perfect supercell.
        'supercell_001', 'supercell_002', ...
            Supercells with displacements
        'primitive':
            Primitive cell.
        'phonon_supercell'
            For phono3py. Perfect supercell for harmonic phonon calculation.
        'phonon_supercell_001', 'phonon_supercell_002', ...
            For phono3py. Supercells with displacements for harmonic phonon
            calculation.

    """
    structures_dict = {}

    digits = len(str(len(ph.supercells_with_displacements)))
    for i, scell in enumerate(ph.supercells_with_displacements):
        structure = phonopy_atoms_to_structure(scell)
        label = "supercell_%s" % str(i + 1).zfill(digits)
        structure.label = "%s %s" % (structure.get_formula(mode="hill_compact"), label)
        structures_dict[label] = structure

    supercell_structure = phonopy_atoms_to_structure(ph.supercell)
    supercell_structure.label = "%s %s" % (
        supercell_structure.get_formula(mode="hill_compact"),
        "supercell",
    )
    structures_dict["supercell"] = supercell_structure

    primitive_structure = phonopy_atoms_to_structure(ph.primitive)
    primitive_structure.label = "%s %s" % (
        primitive_structure.get_formula(mode="hill_compact"),
        "primitive cell",
    )
    structures_dict["primitive"] = primitive_structure

    # phono3py
    if "phonon_supercell_matrix" in ph.__dir__():
        if ph.phonon_supercell_matrix is not None:
            digits = len(str(len(ph.phonon_supercells_with_displacements)))
            for i, scell in enumerate(ph.phonon_supercells_with_displacements):
                structure = phonopy_atoms_to_structure(scell)
                label = "phonon_supercell_%s" % str(i + 1).zfill(digits)
                structure.label = "%s %s" % (
                    structure.get_formula(mode="hill_compact"),
                    label,
                )
                structures_dict[label] = structure
            structure = phonopy_atoms_to_structure(ph.phonon_supercell)
            structure.label = "%s %s" % (
                structure.get_formula(mode="hill_compact"),
                "phonon_supercell",
            )
            structures_dict["phonon_supercell"] = structure

    return structures_dict



import numpy as np
from phonopy import Phonopy
from phonopy.structure.dataset import get_displacements_and_forces
from aiida.engine import WorkChain
from aiida.plugins import WorkflowFactory, DataFactory
from aiida.orm import Bool, Float, Int, QueryBuilder, Group, load_node, Code
from aiida.engine import while_, if_, calcfunction
from aiida_phonopy.common.utils import (
    phonopy_atoms_from_structure,
    get_remote_fc_calculation_settings,
)

Dict = DataFactory("dict")
ArrayData = DataFactory("array")
PhonopyWorkChain = WorkflowFactory("phonopy.phonopy")

"""

Parameters in get_random_displacements and collect_dataset
----------------------------------------------------------
To control how to include previous snapshots, several parameters
can be used.

1) number_of_steps_for_fitting : Int
    Maximum number of previous phonon calculation steps included. When
    number of the previous phonon calculations is smaller than this
    number, all the previous phonon calculations are included. But
    this number is used for (2) in any case.
2) linear_decay : True
    This will be an option, but currently always True.
    This controls weights of previous phonon calculations. One previous
    phonon calculation is included 100%. (number_of_steps_for_fitting + 1)
    previous phonon calculation is not include. Between them, those are
    included with linear scalings. The snapshots in each phonon calculation
    are included from the first element of the list to the specified
    number, i.e., [:num_included_snapshots].
3) include_ratio : Float
    After collecting snapshots in (2), all the snapshots are sorted by
    total energies. Then only lowest energy snapshots with 'include_ratio'
    are included to calculate force constants by fitting using ALM.
4) random_seed : Int
    Using force constants created in (3), phonons are calculated at
    commensurate points, and these phonons are used to generate atomic
    displacements by sampling harmonic oscillator distribution function.
    For this 'random_seed' is used when provided.

"""


def get_random_displacements(
    structure,
    number_of_snapshots,
    temperature,
    phonon_setting_info,
    dataset_for_fc,
    random_seed=None,
):
    """Generate supercells with random displacemens.

    The random displacements are generated from phonons and harmonic
    oscillator distribution function of canonical ensemble. The input
    phonons are calculated from force constants calculated from
    forces and displacemens of the supercell snapshots in previous
    phonon calculation steps.

    Returns
    -------
    dataset : Dict
        Displacement datasets to run force calculations.

    """
    # Calculate force constants by fitting using ALM
    smat = phonon_setting_info["supercell_matrix"]
    ph = Phonopy(
        phonopy_atoms_from_structure(structure),
        supercell_matrix=smat,
        primitive_matrix="auto",
    )
    d = dataset_for_fc.get_array("displacements")
    f = dataset_for_fc.get_array("forces")
    ph.dataset = {"displacements": d, "forces": f}
    ph.produce_force_constants(fc_calculator="alm")

    if random_seed is not None:
        _random_seed = random_seed.value
    else:
        _random_seed = None
    dataset = _generate_random_displacements(
        ph, number_of_snapshots.value, temperature.value, random_seed=_random_seed
    )
    return Dict(dict=dataset)


@calcfunction
def collect_dataset(number_of_steps_for_fitting, include_ratio, linear_decay, **data):
    """Collect supercell displacements, forces, and energies.

    Returns
    -------
    dataset : ArrayData
        Displacements and forces used for calculating force constants
        to generate random displacements.
    supercell_energies : Dict
        'supercell_energies' : List of list
            Total energies of snapshots before step (3) above.
        'included' : List of list
            Finally included snapshots after step (4). The snapshots are
            indexed by concatenating list of list in step (3).

    """
    nitems = max([int(key.split("_")[-1]) for key in data.keys() if "forces" in key])

    forces_in_db = []
    ph_info_in_db = []
    for i in range(nitems):
        force_sets = data["forces_%d" % (i + 1)]
        phonon_setting_info = data["phonon_setting_info_%d" % (i + 1)]
        forces_in_db.append(force_sets)
        ph_info_in_db.append(phonon_setting_info)

    d, f, energies = _extract_dataset_from_db(forces_in_db, ph_info_in_db)

    displacements, forces, included = create_dataset(
        d,
        f,
        energies,
        max_items=number_of_steps_for_fitting.value,
        ratio=include_ratio.value,
        linear_decay=linear_decay.value,
    )
    dataset = ArrayData()
    dataset.set_array("forces", forces)
    dataset.set_array("displacements", displacements)
    supercell_energies = Dict(dict={"energies": energies, "included": included})

    return {"dataset": dataset, "supercell_energies": supercell_energies}


def create_dataset(
    displacements, forces, energies, max_items=None, ratio=None, linear_decay=False
):
    """Collect data for force constants calculation."""
    included = _choose_snapshots_by_linear_decay(
        displacements, forces, max_items=max_items, linear_decay=linear_decay
    )

    # Remove snapshots that have high energies when include_ratio is given.
    if energies is not None and ratio is not None:
        if 0 < ratio and ratio < 1:
            included = _remove_high_energy_snapshots(energies, included, ratio)

    _displacements, _forces, _energies = _include_snapshots(
        displacements, forces, energies, included
    )

    # Concatenate the data
    d = np.concatenate(_displacements, axis=0)
    f = np.concatenate(_forces, axis=0)

    return d, f, included


def _extract_dataset_from_db(forces_in_db, ph_info_in_db):
    nitems = len(forces_in_db)
    displacements = []
    forces = []
    energies = []

    for i in range(nitems):
        force_sets = forces_in_db[i].get_array("force_sets")
        dataset = ph_info_in_db[i]["displacement_dataset"]
        disps, _ = get_displacements_and_forces(dataset)

        forces.append(force_sets)
        if "energies" in forces_in_db[i].get_arraynames():
            energy_sets = forces_in_db[i].get_array("energies")
            energies.append(energy_sets)
        displacements.append(disps)

    return displacements, forces, energies


def _choose_snapshots_by_linear_decay(
    displacements, forces, max_items=None, linear_decay=False
):
    """Choose snapshots by linear_decay.

    With linear_decay=True, numbers of snapshots to be taken
    are biased. Older snapshots are taken lesser. The fraction
    of the number of snapshots in each previous phonon calculation
    (many snapshots in one phonon calculation) to be taken is defined
    linearly by

        ratios = (np.arange(max_items, dtype=float) + 1) / max_items

    where max_items is the number of previous phonon calculations to be
    included at maximum.

    """
    assert len(forces) == len(displacements)

    nitems = len(forces)
    if max_items is None:
        _max_items = nitems
    else:
        _max_items = max_items

    if linear_decay:
        ratios = (np.arange(_max_items, dtype=float) + 1) / _max_items
    else:
        ratios = np.ones(_max_items, dtype=int)
    ratios = ratios[-nitems:]
    included = []

    for i in range(nitems):
        n = len(forces[i])

        assert n == len(displacements[i])

        n_in = int(ratios[i] * n + 0.5)
        if n < n_in:
            n_in = n
        included.append(
            [
                True,
            ]
            * n_in
            + [
                False,
            ]
            * (n - n_in)
        )

    return included


def _include_snapshots(displacements, forces, energies, included):
    _forces = [forces[i][included_batch] for i, included_batch in enumerate(included)]
    _displacements = [
        np.array(displacements[i])[included_batch]
        for i, included_batch in enumerate(included)
    ]
    if energies:
        _energies = [
            energies[i][included_batch] for i, included_batch in enumerate(included)
        ]
    else:
        _energies = []
    return _displacements, _forces, _energies


def _remove_high_energy_snapshots(energies, included, ratio):
    """Reject high energy snapshots.

    Parameters
    ----------
    energies : list of ndarray
        List of supercell total energies in each batch
    included : list of list of bool
        List of list of True/False as included snapshots in each batch
        Rejected elements are turned to False.
    ratio : float
        How much ratio of lowest energy snapshots is included after
        sorting by energy.

    Returns
    -------
    ret_included :
        Rejected snapshots are turned to False from 'included'.

    """
    concat_included = np.concatenate(included)
    concat_energies = np.concatenate(energies)
    included_energies = concat_energies[concat_included]
    included_indices = np.arange(len(concat_included))[concat_included]

    num_include = int(ratio * len(included_energies) + 0.5)
    if len(included_energies) < num_include:
        num_include = len(included_energies)
    _indices = np.argsort(included_energies)[:num_include]
    included_indices_after_energy = included_indices[_indices]

    bool_list = [
        False,
    ] * len(concat_included)
    for i in included_indices_after_energy:
        bool_list[i] = True
    ret_included = []
    count = 0
    for included_batch in included:
        ret_included.append(bool_list[count : (count + len(included_batch))])
        count += len(included_batch)
    return ret_included


def _modify_force_constants(ph):
    """Apply treatment to imaginary modes.

    This method modifies force constants to make phonon frequencies
    be real from imaginary. This treatment is expected to be finally
    forgotten after many iterations. Therefore it is unnecessary
    to be physical and can be physically dirty. If it works, it is OK,
    though good treatment may contribute to quick convergence.

    1) All frequencies at commensurate points are converted to their
       absolute values. freqs -> |freqs|.
    2) Phonon frequencies in the interval 0.01 < |freqs| < 0.5
       are shifted by |freqs| + 1.

    """
    # temperature=300 is used just to invoke this feature.
    ph.generate_displacements(number_of_snapshots=1, temperature=300)
    rd = ph.random_displacements
    freqs = np.abs(rd.frequencies)
    condition = np.logical_and(freqs > 0.01, freqs < 0.5)
    rd.frequencies = np.where(condition, freqs + 1, freqs)
    rd.run_d2f()
    ph.force_constants = rd.force_constants


def _generate_random_displacements(
    ph, number_of_snapshots, temperature, random_seed=None
):
    # Treatment of imaginary modes
    _modify_force_constants(ph)

    ph.generate_displacements(
        number_of_snapshots=number_of_snapshots,
        random_seed=random_seed,
        temperature=temperature,
    )

    return ph.dataset


class IterHarmonicApprox(WorkChain):
    """Workchain for harmonic force constants by iterative approach.

    By default, the calculation starts with normal phonon calculation,
    i.e., in this context, which corresponds to roughly 0K force constants.
    Then the iteration loop starts. The first run is the iteration-1.
    The iteration stops after finishing that of max_iteration. Each phonon
    calculation is named 'step'.

    Steps
    -----
    0. Initial phonon calculation at 0K
    1. First phonon calculation at specified temperature. Random
       displacements are created from step-0.
    2. Second phonon calculation at specified temperature. Random
       displacements are created from step-1.
    3. Third phonon calculation at specified temperature. Random
       displacements are created from steps 1 and 2 if
       number_of_snapshots >= 2. Otherwise only the result from
       step-2 is used.
    4. Four phonon calculation at specified temperature. Random
       displacements are created from number_of_snapshots previous
       existing steps excluding step-0.
    *. Continue until iteration number = max_iteration number.

    Manual termination of iteration loop
    ------------------------------------
    It is possible to terminate at the initial point of each iteration.
    This option is not very recommended to use because not reproducible
    mechanically, but can be useful for experimental calculation.

    This is achieved by just creating AiiDA Group whose label is its
    uuid string that ejected by AiiDA, i.e., self.uuid.

    inputs
    ------
    Most of inputs are imported from PhonopyWorkChain. Specific inputs
    of this workchain are as follows:

    max_iteration : Int
        Maximum number of iterations.
    number_of_snapshots : Int
        Number of generated supercell snapshots with random displacements
        at a temperature.
    number_of_steps_for_fitting : Int
        Displacements and respective forces of supercells in the previous
        number_of_steps_for_fitting are used to simultaneously fit to
        force constants.
    temperature : Float
        Temperature (K).
    include_ratio : Float
        How much supercell forces are included from lowest supercell energies.
        Default is 1.0.
    lienar_decay : Bool
        This controls weights of previous phonon calculations. One previous
        phonon calculation is included 100%. (number_of_steps_for_fitting + 1)
        previous phonon calculation is not include. Between them, those are
        included with linear scalings. The snapshots in each phonon calculation
        are included from the first element of the list to the specified
        number, i.e., [:num_included_snapshots]. Default is False.
    random_seed : Int, optional
        Random seed used to sample in canonical ensemble harmonic oscillator
        space. The value must be 32bit unsigned int. Unless specified,
        random seed will not be fixed.
    initial_nodes : Dict, optional
        This gives the initial nodes that contain sets of forces, which are
        provided by PKs or UUIDs.

    """

    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.expose_inputs(
            PhonopyWorkChain,
            exclude=["immigrant_calculation_folders", "calculation_nodes", "dry_run"],
        )
        spec.input("max_iteration", valid_type=Int, default=lambda: Int(10))
        spec.input("number_of_snapshots", valid_type=Int, default=lambda: Int(100))
        spec.input(
            "number_of_steps_for_fitting", valid_type=Int, default=lambda: Int(4)
        )
        spec.input("temperature", valid_type=Float, default=lambda: Float(300.0))
        spec.input("include_ratio", valid_type=Float, default=lambda: Float(1))
        spec.input("linear_decay", valid_type=Bool, default=lambda: Bool(False))
        spec.input("random_seed", valid_type=Int, required=False)
        spec.input("initial_nodes", valid_type=Dict, required=False)
        spec.outline(
            cls.initialize,
            if_(cls.import_initial_nodes)(cls.set_initial_nodes,).else_(
                cls.run_initial_phonon,
            ),
            while_(cls.is_loop_finished)(
                cls.collect_displacements_and_forces,
                if_(cls.remote_phonopy)(
                    cls.run_force_constants_calculation_remote,
                    cls.generate_displacements,
                ).else_(
                    cls.generate_displacements_local,
                ),
                cls.run_phonon,
            ),
            cls.finalize,
        )

    def remote_phonopy(self):
        return self.inputs.remote_phonopy

    def import_initial_nodes(self):
        return "initial_nodes" in self.inputs

    def initialize(self):
        self.report("initialize (%s)" % self.uuid)
        self.ctx.iteration = 0
        self.ctx.prev_nodes = []
        if self.remote_phonopy():
            self.ctx.ph_settings = get_remote_fc_calculation_settings(
                self.inputs.phonon_settings
            )

    def is_loop_finished(self):
        qb = QueryBuilder()
        qb.append(Group, filters={"label": {"==": self.uuid}})
        if qb.count() == 1:
            self.report(
                "Iteration loop is manually terminated at step %d." % self.ctx.iteration
            )
            return False

        self.ctx.iteration += 1
        return self.ctx.iteration <= self.inputs.max_iteration.value

    def set_initial_nodes(self):
        self.report("set_initial_phonon")
        node_ids = self.inputs.initial_nodes["nodes"]
        self.ctx.prev_nodes = [load_node(node_id) for node_id in node_ids]

    def run_initial_phonon(self):
        self.report("run_initial_phonon")
        inputs = self._get_phonopy_inputs(None, None)
        inputs["metadata"].label = "Initial phonon calculation"
        inputs["metadata"].description = "Initial phonon calculation"
        future = self.submit(PhonopyWorkChain, **inputs)
        self.ctx.initial_node = future
        self.report("{} pk = {}".format(inputs["metadata"].label, future.pk))
        self.to_context(**{"phonon_0": future})

    def collect_displacements_and_forces(self):
        self.report("run_generate_displacements_%d" % self.ctx.iteration)

        num_batches = self.inputs.number_of_steps_for_fitting.value

        # Initial nodes are not specified, 0K phonon is in
        # self.ctx.initial_node. This is only once used to generate random
        # displacements. In the following steps, phonons calculated at
        # specified temperature are used to generate random displacements.
        if len(self.ctx.prev_nodes) == 0:
            nodes = [
                self.ctx.initial_node,
            ]
        elif len(self.ctx.prev_nodes) < num_batches:
            nodes = self.ctx.prev_nodes
        else:
            nodes = self.ctx.prev_nodes[-num_batches:]
        data_for_fc = {}
        for i, node in enumerate(nodes):
            data_for_fc["forces_%d" % (i + 1)] = node.outputs.force_sets
            ph_info = node.outputs.phonon_setting_info
            data_for_fc["phonon_setting_info_%d" % (i + 1)] = ph_info
        ret_Dict = collect_dataset(
            self.inputs.number_of_steps_for_fitting,
            self.inputs.include_ratio,
            self.inputs.linear_decay,
            **data_for_fc
        )
        self.ctx.dataset_for_fc = ret_Dict["dataset"]

    def generate_displacements_local(self):
        if "random_seed" in self.inputs:
            data_rd = {"random_seed": self.inputs.random_seed}
        else:
            data_rd = {}
        dataset = get_random_displacements(
            self.inputs.structure,
            self.inputs.number_of_snapshots,
            self.inputs.temperature,
            self.inputs.phonon_settings,
            self.ctx.dataset_for_fc,
            **data_rd
        )
        self.ctx.dataset = dataset

    def run_force_constants_calculation_remote(self):
        """Run force constants calculation by PhonopyCalculation."""
        self.report("remote force constants calculation %d" % self.ctx.iteration)

        code_string = self.inputs.code_string.value
        builder = Code.get_from_string(code_string).get_builder()
        builder.structure = self.inputs.structure
        builder.settings = self.ctx.ph_settings
        builder.metadata.options.update(self.inputs.phonon_settings["options"])
        builder.metadata.label = "Force constants calculation %d" % self.ctx.iteration
        builder.dataset = self.ctx.dataset_for_fc
        builder.fc_only = Bool(True)
        future = self.submit(builder)

        self.report("Force constants remote calculation: {}".format(future.pk))
        label = "force_constants_%d" % self.ctx.iteration
        self.to_context(**{label: future})

    def generate_displacements(self):
        label = "force_constants_%d" % self.ctx.iteration
        fc_array = self.ctx[label].outputs.force_constants
        fc = fc_array.get_array("force_constants")
        phonon_setting_info = self.inputs.phonon_settings
        smat = phonon_setting_info["supercell_matrix"]
        ph = Phonopy(
            phonopy_atoms_from_structure(self.inputs.structure),
            supercell_matrix=smat,
            primitive_matrix="auto",
        )
        ph.force_constants = fc

        if "random_seed" in self.inputs:
            random_seed = self.inputs.random_seed.value
        else:
            random_seed = None
        dataset = _generate_random_displacements(
            ph,
            self.inputs.number_of_snapshots.value,
            self.inputs.temperature.value,
            random_seed=random_seed,
        )
        self.ctx.dataset = Dict(dict=dataset)

    def run_phonon(self):
        self.report("run_phonon_%d" % self.ctx.iteration)

        inputs = self._get_phonopy_inputs(self.ctx.dataset, False)
        label = "Phonon calculation %d" % self.ctx.iteration
        inputs["metadata"].label = label
        inputs["metadata"].description = label
        future = self.submit(PhonopyWorkChain, **inputs)
        self.ctx.prev_nodes.append(future)
        self.report("{} pk = {}".format(label, future.pk))
        label = "phonon_%d" % self.ctx.iteration
        self.to_context(**{label: future})

    def finalize(self):
        self.report("IterHarmonicApprox finished at %d" % (self.ctx.iteration - 1))

    def _get_phonopy_inputs(self, dataset, is_nac):
        inputs_in = self.exposed_inputs(PhonopyWorkChain)
        inputs = inputs_in.copy()
        phonon_settings = inputs_in["phonon_settings"].get_dict()
        if is_nac is not None:
            phonon_settings["is_nac"] = is_nac
        inputs["phonon_settings"] = Dict(dict=phonon_settings)
        if dataset is not None:
            inputs["displacement_dataset"] = dataset
        return inputs



"""CalcJob to run phonopy at a remote host."""

from aiida.plugins import DataFactory
from aiida_phonopy.calcs.base import BasePhonopyCalculation
from aiida.common import InputValidationError
from aiida_phonopy.common.file_generators import (
    get_FORCE_SETS_txt,
    get_phonopy_options,
    get_phonopy_yaml_txt,
)


BandsData = DataFactory("array.bands")
ArrayData = DataFactory("array")
XyData = DataFactory("array.xy")
Dict = DataFactory("dict")
Str = DataFactory("str")


class PhonopyCalculation(BasePhonopyCalculation):
    """Phonopy calculation."""

    _OUTPUT_PROJECTED_DOS = "projected_dos.dat"
    _OUTPUT_TOTAL_DOS = "total_dos.dat"
    _OUTPUT_THERMAL_PROPERTIES = "thermal_properties.yaml"
    _OUTPUT_BAND_STRUCTURE = "band.yaml"
    _INOUT_FORCE_CONSTANTS = "force_constants.hdf5"
    _INPUT_CELL = "phonopy_cells.yaml"
    _INPUT_FORCE_SETS = "FORCE_SETS"

    @classmethod
    def define(cls, spec):
        """Define inputs, outputs, and outline."""
        super().define(spec)

        # parser_name has to be set to invoke parsing.
        spec.inputs["metadata"]["options"]["parser_name"].default = "phonopy"
        spec.inputs["metadata"]["options"]["output_filename"].default = "phonopy.yaml"

        spec.output(
            "force_constants",
            valid_type=ArrayData,
            required=False,
            help="Calculated force constants",
        )
        spec.output(
            "dos", valid_type=XyData, required=False, help="Calculated total DOS"
        )
        spec.output(
            "pdos", valid_type=XyData, required=False, help="Calculated projected DOS"
        )
        spec.output(
            "thermal_properties",
            valid_type=XyData,
            required=False,
            help="Calculated thermal properties",
        )
        spec.output(
            "band_structure",
            valid_type=BandsData,
            required=False,
            help="Calculated phonon band structure",
        )
        spec.output("version", valid_type=Str, required=False, help="Version number")

    def prepare_for_submission(self, folder):
        """Prepare calcinfo."""
        return super().prepare_for_submission(folder)

    def _create_additional_files(self, folder):
        self.logger.info("create_additional_files")

        self._create_phonopy_yaml(folder)
        self._create_FORCE_SETS(folder)
        mesh_opts, fc_opts = get_phonopy_options(
            self.inputs.settings["postprocess_parameters"]
        )

        self._internal_retrieve_list = [
            self._INOUT_FORCE_CONSTANTS,
            self.inputs.metadata.options.output_filename,
        ]
        self._additional_cmd_params = [
            ["--writefc", "--writefc-format=hdf5"] + fc_opts,
            ["--readfc", "--readfc-format=hdf5"],
            ["--readfc", "--readfc-format=hdf5"],
        ]

        # First run with --writefc, and with --readfc for remaining runs
        if self.inputs.fc_only:
            self._calculation_cmd = [
                ["-c", self._INPUT_CELL],
            ]
        else:
            self._calculation_cmd = [
                ["-c", self._INPUT_CELL, "--pdos=auto"] + mesh_opts,
                ["-c", self._INPUT_CELL, "-t", "--dos"] + mesh_opts,
                [
                    "-c",
                    self._INPUT_CELL,
                    "--band=auto",
                    "--band-points=101",
                    "--band-const-interval",
                ],
            ]
            self._internal_retrieve_list += [
                self._OUTPUT_TOTAL_DOS,
                self._OUTPUT_PROJECTED_DOS,
                self._OUTPUT_THERMAL_PROPERTIES,
                self._OUTPUT_BAND_STRUCTURE,
            ]

    def _create_phonopy_yaml(self, folder):
        phpy_yaml_txt = get_phonopy_yaml_txt(
            self.inputs.structure,
            supercell_matrix=self.inputs.settings["supercell_matrix"],
        )
        with folder.open(self._INPUT_CELL, "w", encoding="utf8") as handle:
            handle.write(phpy_yaml_txt)

    def _create_FORCE_SETS(self, folder):
        if "force_sets" in self.inputs:
            force_sets = self.inputs.force_sets
        else:
            force_sets = None
        if "displacement_dataset" in self.inputs.settings.keys():
            dataset = self.inputs.settings["displacement_dataset"]
        elif "dataset" in self.inputs.settings.keys():
            dataset = self.inputs.settings["dataset"]
        elif (
            "dataset" in self.inputs
            and "displacements" in self.inputs.dataset.get_arraynames()
        ):
            dataset = {"displacements": self.inputs.dataset.get_array("displacements")}
            if "forces" in self.inputs.dataset.get_arraynames():
                dataset["forces"] = self.inputs.dataset.get_array("forces")
                if force_sets is not None:
                    force_sets = None
        else:
            dataset = None

        # can work both for type-I and type-II
        force_sets_txt = get_FORCE_SETS_txt(dataset, force_sets=force_sets)
        if force_sets_txt is None:
            msg = "Displacements or forces were not found."
            raise InputValidationError(msg)

        with folder.open(self._INPUT_FORCE_SETS, "w", encoding="utf8") as handle:
            handle.write(force_sets_txt)



from setuptools import setup, find_packages

import json

if __name__ == "__main__":
    with open("setup.json", "r") as info:
        kwargs = json.load(info)

    setup(
        reentry_register=True,
        include_package_data=True,
        packages=find_packages(exclude=["aiida"]),
        **kwargs
    )



"""Workflow to calculate supercell forces."""

import numpy as np
from aiida.engine import WorkChain, calcfunction, if_
from aiida.orm import Code
from aiida.plugins import DataFactory, WorkflowFactory

from aiida_phonopy.common.builders import (
    get_calculator_process,
    get_vasp_immigrant_inputs,
    get_workchain_inputs,
)
from aiida_phonopy.common.utils import (
    compare_structures,
    get_structure_from_vasp_immigrant,
)

Float = DataFactory("float")
Dict = DataFactory("dict")
Str = DataFactory("str")
ArrayData = DataFactory("array")
StructureData = DataFactory("structure")


def _get_forces(outputs, code_string):
    """Return supercell force ArrayData."""
    code = Code.get_from_string(code_string)
    plugin_name = code.get_input_plugin_name()
    if plugin_name == "vasp.vasp":
        if "forces" in outputs and "final" in outputs.forces.get_arraynames():
            forces_data = get_vasp_forces(outputs.forces)
        else:
            return None
    elif plugin_name == "quantumespresso.pw":
        if (
            "output_trajectory" in outputs
            and "forces" in outputs.output_trajectory.get_arraynames()
        ):
            forces_data = get_qe_forces(outputs.output_trajectory)
        else:
            return None
    return forces_data


@calcfunction
def get_vasp_forces(forces):
    """Return VASP forces ArrayData."""
    forces_data = ArrayData()
    forces_data.set_array("forces", forces.get_array("final"))
    forces_data.label = "forces"
    return forces_data


@calcfunction
def get_qe_forces(output_trajectory):
    """Return QE forces ArrayData."""
    forces_data = ArrayData()
    forces_data.set_array("forces", output_trajectory.get_array("forces")[-1])
    forces_data.label = "forces"
    return forces_data


def _get_energy(outputs, code_string):
    """Return supercell energy ArrayData."""
    code = Code.get_from_string(code_string)
    plugin_name = code.get_input_plugin_name()
    if plugin_name == "vasp.vasp":
        ekey = "energy_extrapolated"
        if "energies" in outputs and ekey in outputs.energies.get_arraynames():
            energy_data = get_vasp_energy(outputs.energies)
        else:
            return None
    elif plugin_name == "quantumespresso.pw":
        if (
            "output_parameters" in outputs
            and "energy" in outputs.output_parameters.keys()
        ):
            energy_data = get_qe_energy(outputs.output_parameters)
    return energy_data


@calcfunction
def get_vasp_energy(energies):
    """Return VASP energy ArrayData.

    energies is an 1D-array of energies.

    {'electronic_step_energies': True} gives SC energies.
    {'electronic_step_energies': False} gives last SC energy.

    """
    energy_data = ArrayData()
    ekey = "energy_extrapolated"
    energies = energies.get_array(ekey)
    energy_data.set_array("energy", np.array(energies, dtype=float))
    energy_data.label = "energy"
    return energy_data


@calcfunction
def get_qe_energy(output_parameters):
    """Return VASP energy ArrayData."""
    energy_data = ArrayData()
    energy_data.set_array(
        "energy",
        np.array(
            [
                output_parameters["energy"],
            ],
            dtype=float,
        ),
    )
    energy_data.label = "energy"
    return energy_data


class ForcesWorkChain(WorkChain):
    """Wrapper to compute supercell forces."""

    @classmethod
    def define(cls, spec):
        """Define inputs, outputs, and outline."""
        super().define(spec)
        spec.input("structure", valid_type=StructureData, required=True)
        spec.input("calculator_inputs", valid_type=dict, required=True, non_db=True)
        spec.input("symmetry_tolerance", valid_type=Float, default=lambda: Float(1e-5))
        spec.input("immigrant_calculation_folder", valid_type=Str, required=False)
        spec.outline(
            if_(cls.import_calculation_from_files)(
                cls.read_calculation_from_folder,
                cls.validate_imported_structure,
            ).else_(
                cls.run_calculation,
            ),
            cls.finalize,
        )

        spec.output("forces", valid_type=ArrayData, required=True)
        spec.output("energy", valid_type=ArrayData, required=False)

        spec.exit_code(
            1001,
            "ERROR_NO_FORCES",
            message="forces could not be retrieved from calculaton.",
        )
        spec.exit_code(
            1002,
            "ERROR_NO_ENERGY",
            message="energy could not be retrieved from calculaton.",
        )
        spec.exit_code(
            1003,
            "ERROR_STRUCTURE_VALIDATION",
            message="input and imported structures are different.",
        )

    def import_calculation_from_files(self):
        """Return boolen for outline."""
        return "immigrant_calculation_folder" in self.inputs

    def run_calculation(self):
        """Run supercell force calculation."""
        self.report("calculate supercell forces")
        process_inputs = get_workchain_inputs(
            self.inputs.calculator_inputs,
            self.inputs.structure,
            label=self.metadata.label,
        )
        CalculatorProcess = get_calculator_process(
            self.inputs.calculator_inputs["code_string"]
        )
        future = self.submit(CalculatorProcess, **process_inputs)
        self.report("{} pk = {}".format(self.metadata.label, future.pk))
        self.to_context(**{"calc": future})

    def read_calculation_from_folder(self):
        """Import supercell force calculation using immigrant."""
        self.report("import supercell force calculation data in files.")
        force_folder = self.inputs.immigrant_calculation_folder
        inputs = get_vasp_immigrant_inputs(
            force_folder.value,
            self.inputs.calculator_inputs.dict,
            label=self.metadata.label,
        )
        VaspImmigrant = WorkflowFactory("vasp.immigrant")
        future = self.submit(VaspImmigrant, **inputs)
        self.report("{} pk = {}".format(self.metadata.label, future.pk))
        self.to_context(**{"calc": future})

    def validate_imported_structure(self):
        """Validate imported supercell structure."""
        self.report("validate imported supercell structures")
        supercell_ref = self.inputs.structure
        supercell_calc = get_structure_from_vasp_immigrant(self.ctx.calc)
        if not compare_structures(
            supercell_ref, supercell_calc, self.inputs.symmetry_tolerance
        ):
            return self.exit_codes.ERROR_STRUCTURE_VALIDATION

    def finalize(self):
        """Finalize force calculation."""
        outputs = self.ctx.calc.outputs
        self.report("create forces ArrayData")
        forces = _get_forces(outputs, self.inputs.calculator_inputs["code_string"])
        if forces is None:
            return self.exit_codes.ERROR_NO_FORCES
        else:
            self.out("forces", forces)

        self.report("create energy ArrayData")
        energy = _get_energy(outputs, self.inputs.calculator_inputs["code_string"])
        if energy is None:
            return self.exit_codes.ERROR_NO_ENERGY
        else:
            self.out("energy", energy)
